import itertools
import textwrap
import warnings
from typing import List, Dict, Optional, Union, Tuple, Any

import qcodes.utils.validators as vals
from qcodes import VisaInstrument
from qcodes.utils.helpers import create_on_off_val_mapping


class UnknownOrEmptySlot(Exception):
    pass


class InvalidValue(Exception):
    pass


class Keithley_3706A(VisaInstrument):
    """
    This is the QCoDeS instrument driver for the Keithley 3706A-SNFP
    System Switch.
    """

    def __init__(self, name: str, address: str,
                 terminator: str = '\n', **kwargs: Any) -> None:
        """
        Args:
            name: Name to use internally in QCoDeS
            address: VISA resource address
        """
        super().__init__(name, address, terminator=terminator, **kwargs)

        self.add_parameter('channel_connect_rule',
                           get_cmd=self._get_channel_connect_rule,
                           set_cmd=self._set_channel_connect_rule,
                           docstring=textwrap.dedent("""\
                                    Controls the connection rule for closing
                                    and opening channels when using
                                    `exclusive_close` and `exclusive_slot_close`
                                    parameters.

                                    If it is set to break before make,
                                    it is ensured that all channels open
                                    before any channels close.

                                    If it is set to make before break, it is
                                    ensured that all channels close before any
                                    channels open.

                                    If it is off, channels open and close
                                    simultaneously."""),
                           vals=vals.Enum('BREAK_BEFORE_MAKE',
                                          'MAKE_BEFORE_BREAK',
                                          'OFF'))

        self.add_parameter('gpib_enabled',
                           get_cmd=self._get_gpib_status,
                           set_cmd=self._set_gpib_status,
                           docstring='Enables or disables GPIB connection.',
                           val_mapping=create_on_off_val_mapping(on_val='true',
                                                                 off_val='false'
                                                                 ))

        self.add_parameter('gpib_address',
                           get_cmd=self._get_gpib_address,
                           get_parser=int,
                           set_cmd=self._set_gpib_address,
                           docstring='Sets and gets the GPIB address.',
                           vals=vals.Ints(1, 30))

        self.add_parameter('lan_enabled',
                           get_cmd=self._get_lan_status,
                           set_cmd=self._set_lan_status,
                           docstring='Enables or disables LAN connection.',
                           val_mapping=create_on_off_val_mapping(on_val='true',
                                                                 off_val='false'
                                                                 ))

        self.connect_message()

    def reset_channel(self, val: str) -> None:
        """
        Resets the specified channels to factory default settings.

        Args:
            val: A string representing the channels, channel ranges,
                backplane relays, slots or channel patterns to be queried.
        """
        if not self._validator(val):
            raise InvalidValue(f'{val} is not a valid specifier. '
                               'The specifier should be channels, channel '
                               'ranges, slots, backplane relays or "allslots".')
        self.write(f"channel.reset('{val}')")

    def open_channel(self, val: str) -> None:
        """
        Opens the specified channels and backplane relays.

        Args:
            val: A string representing the channels, channel ranges,
                backplane relays, slots or channel patterns to be queried.
        """

        if not self._validator(val):
            raise InvalidValue(f'{val} is not a valid specifier. '
                               'The specifier should be channels, channel '
                               'ranges, slots, backplane relays or "allslots".')
        self.write(f"channel.open('{val}')")

    def close_channel(self, val: str) -> None:
        """
        Closes the channels and backplane relays.

        Args:
            val: A string representing the channels, channel ranges,
                backplane relays to be queried.
        """
        slots = ['allslots', *self._get_slot_names()]
        forbidden_channels = self.get_forbidden_channels("allslots")
        if val in slots:
            raise InvalidValue("Slots cannot be closed all together.")
        if not self._validator(val):
            raise InvalidValue(f'{val} is not a valid specifier. '
                               'The specifier should be channels or channel '
                               'ranges and associated backplane relays.')
        if val in forbidden_channels.split(','):
            warnings.warn('You are attempting to close channels that are '
                          'forbidden to close.', UserWarning, 2)

        self._warn_on_disengaged_interlocks(val)

        self.write(f"channel.close('{val}')")

    def _warn_on_disengaged_interlocks(self, val: str) -> None:
        """
        Checks if backplance channels among the given specifiers can be
        energized dependening on respective hardware interlocks being
        engaged, and raises a warning for those backplane channels which
        cannot be energized.

        Args:
            val: A string representing the channels, channel ranges,
                backplane relays.
        """
        states = self.get_interlock_state()
        val_specifiers = val.split(",")
        for channel in val_specifiers:
            if self._is_backplane_channel(channel):
                slot = channel[0]
                interlock_state = [
                    state for state in states if state["slot_no"] == slot
                ][0]
                if interlock_state["state"] == "Interlock is disengaged":
                    warnings.warn(
                        f"The hardware interlock in Slot "
                        f'{interlock_state["slot_no"]} is disengaged. '
                        f"The analog backplane relay {channel} "
                        "cannot be energized.",
                        UserWarning,
                        2,
                    )

    def _is_backplane_channel(self, channel_id: str) -> bool:
        if len(channel_id) != 4:
            raise InvalidValue(f"{channel_id} is not a valid channel id")
        if channel_id[1] == "9":
            return True
        return False

    def exclusive_close(self, val: str) -> None:
        """
        Closes the specified channels such that any presently closed channels
        opens if they are not in the specified by the parameter.

        Args:
            val: A string representing the channels, channel ranges,
                backplane relays to be queried.
        """
        slots = ['allslots', *self._get_slot_names()]
        if val in slots:
            raise InvalidValue("Slots cannot be exclusively closed.")
        if val == "":
            raise InvalidValue('An empty string may cause all channels and '
                               'associated backplane relays to open. Use '
                               '"open_channel" parameter instead.')
        if not self._validator(val):
            raise InvalidValue(f'{val} is not a valid specifier. '
                               'The specifier should be channels or channel '
                               'ranges and associated backplane relays.')
        self.write(f"channel.exclusiveclose('{val}')")

    def exclusive_slot_close(self, val: str) -> None:
        """
        Closes the specified channels on the associated slots abd opens any
        other channels if they are not specified by the parameter.

        Args:
            val: A string representing the channels, channel ranges,
                backplane relays to be queried.
        """
        slots = ['allslots', *self._get_slot_names()]
        if val in slots:
            raise InvalidValue("Slots cannot be exclusively closed.")
        if val == "":
            raise InvalidValue('Argument cannot be an empty string.')
        if not self._validator(val):
            raise InvalidValue(f'{val} is not a valid specifier. '
                               'The specifier should be channels or channel '
                               'ranges and associated backplane relays.')
        self.write(f"channel.exclusiveslotclose('{val}')")

    def _get_channel_connect_rule(self) -> str:
        connect_rule = {1: 'BREAK_BEFORE_MAKE',
                        2: 'MAKE_BEFORE_BREAK',
                        0: 'OFF'}
        rule = self.ask('channel.connectrule')
        return connect_rule[int(float(rule))]

    def _set_channel_connect_rule(self, val: str) -> None:
        self.write(f'channel.connectrule = channel.{val}')

    def _get_gpib_status(self) -> str:
        return self.ask('comm.gpib.enable')

    def _set_gpib_status(self, val: Union[str, bool]) -> None:
        self.write(f'comm.gpib.enable = {val}')

    def _get_lan_status(self) -> str:
        return self.ask('comm.lan.enable')

    def _set_lan_status(self, val: Union[str, bool]) -> None:
        self.write(f'comm.lan.enable = {val}')

    def _get_gpib_address(self) -> int:
        return int(float(self.ask('gpib.address')))

    def _set_gpib_address(self, val: int) -> None:
        self.write(f'gpib.address = {val}')

    def get_closed_channels(self, val: str) -> Optional[List[str]]:
        """
        Queries for the closed channels.

        Args:
            val: A string representing the channels,
                backplane relays or channel patterns to be queried.
        """
        if val == "":
            raise InvalidValue('Argument cannot be an empty string.')
        if not self._validator(val):
            raise InvalidValue(f'{val} is not a valid specifier. '
                               'The specifier should be channels, channel '
                               'ranges, slots, backplane relays or "allslots".')
        data = self.ask(f"channel.getclose('{val}')")
        if data == 'nil':
            return None
        return data.split(';')

    def set_forbidden_channels(self, val: str) -> None:
        """
        Prevents the closing of specified channels and backplane
        relays.

        Args:
            val: A string representing channels and backplane relays
                to make forbidden to close.
        """
        if not self._validator(val):
            raise InvalidValue(f'{val} is not a valid specifier. '
                               'The specifier should be channels, channel '
                               'ranges, slots, backplane relays or "allslots".')
        self.write(f"channel.setforbidden('{val}')")

    def get_forbidden_channels(self, val: str) -> str:
        """
        Returns a string that lists the channels and backplane relays
        that are forbidden to close.

        Args:
            val: A string representing the channels,
                backplane relays or channel patterns to be queried to see
                if they are forbidden to close.
        """
        if not self._validator(val):
            raise InvalidValue(f'{val} is not a valid specifier. '
                               'The specifier should be channels, channel '
                               'ranges, slots, backplane relays or "allslots".')
        return self.ask(f"channel.getforbidden('{val}')")

    def clear_forbidden_channels(self, val: str) -> None:
        """
        Clears the list of channels that are forbidden to close.

        Args:
            val: A string representing the channels that will no longer
                be listed as forbidden to close.
        """
        if not self._validator(val):
            raise InvalidValue(f'{val} is not a valid specifier. '
                               'The specifier should be channels, channel '
                               'ranges, slots, backplane relays or "allslots".')
        self.write(f"channel.clearforbidden('{val}')")

    def set_delay(self, val: str, delay_time: float) -> None:
        """
        Sets an additional delay time for the specified channels.

        Args:
            val: A string representing the channels for which there will
                be an additional delay time.
            delay_time: Delay time for the specified channels in seconds.
        """
        backplanes = self.get_analog_backplane_specifiers()
        specifiers = val.split(',')
        for element in specifiers:
            if element in backplanes:
                raise InvalidValue("Additional delay times cannot be set for "
                                   "analog backplane relays.")
        if not self._validator(val):
            raise InvalidValue(f'{val} is not a valid specifier. '
                               'The specifier should be channels, channel '
                               'ranges, slots, or "allslots".')
        self.write(f"channel.setdelay('{val}', {delay_time})")

    def get_delay(self, val: str) -> List[float]:
        """
        Queries for the delay times.

        Args:
            val: A string representing the channels to query for
                additional delay times.
        """
        backplanes = self.get_analog_backplane_specifiers()
        specifiers = val.split(',')
        for element in specifiers:
            if element in backplanes:
                raise InvalidValue("Additional delay times cannot be set for "
                                   "analog backplane relays.")
        if not self._validator(val):
            raise InvalidValue(f'{val} is not a valid specifier. '
                               'The specifier should be channels, channel '
                               'ranges, slots, or "allslots".')
        delay_times = [float(x) for x in self.ask(
                       f"channel.getdelay('{val}')").split(',')]
        return delay_times

    def set_backplane(self, val: str, backplane: str) -> None:
        """
        Sets the analog backplane relays to use with given channels
        when they are used in switching applications.

        Args:
            val: A string representing the list of channels to change.
            backplane: A string representing the list of analog backplane
                relays to set for the channels specified.
        """
        states = self.get_interlock_state()
        backplanes = self.get_analog_backplane_specifiers()
        plane_specifiers = backplane.split(',')
        val_specifiers = val.split(',')
        for element in states:
            if element['state'] == 'Interlock is disengaged':
                warnings.warn(f'The hardware interlock in Slot '
                              f'{element["slot_no"]} is disengaged. '
                              'The corresponding analog backplane relays '
                              'cannot be energized.', UserWarning, 2)
        for elem in val_specifiers:
            if elem in backplanes:
                raise InvalidValue(f'{val} is not a valid specifier. '
                                   'The specifier cannot be analog '
                                   'backplane relay.')
        if not self._validator(val):
            raise InvalidValue(f'{val} is not a valid specifier. '
                               'The specifier should be channels, channel '
                               'ranges, slots, or "allslots".')
        for plane in plane_specifiers:
            if plane not in backplanes:
                raise InvalidValue(f'{backplane} is not a valid specifier. '
                                   'The specifier should be analog '
                                   'backplane relay.')
        self.write(f"channel.setbackplane('{val}', '{backplane}')")

    def get_backplane(self, val: str) -> str:
        """
        Lists the backplane relays that are controlled in switching
        applications for specified channels.

        Args:
            val: A string representing the channels being queried.
        """
        backplanes = self.get_analog_backplane_specifiers()
        specifiers = val.split(',')
        for element in specifiers:
            if element in backplanes:
                raise InvalidValue(f'{val} cannot be a analog backplane relay.')
        if not self._validator(val):
            raise InvalidValue(f'{val} is not a valid specifier. '
                               'The specifier should be channels, channel '
                               'ranges, slots, or "allslots".')
        return self.ask(f"channel.getbackplane('{val}')")

    def _get_slot_ids(self) -> List[str]:
        """
        Returns the slot ids of the installed cards.
        """
        cards = self.get_switch_cards()
        slot_id = [f"{card['slot_no']}" for card in cards]
        return slot_id

    def _get_slot_names(self) -> List[str]:
        """
        Returns the names of the slots as "slotX",
        where "X" is the slot id.
        """
        slot_id = self._get_slot_ids()
        slot_names = [f'slot{x}' for x in slot_id]
        return slot_names

    def _get_number_of_rows(self) -> List[int]:
        """
        Returns the total number of rows of the installed cards.
        """
        slot_id = self._get_slot_ids()
        total_number_of_rows = [int(float(self.ask(
            f'slot[{i}].rows.matrix')))
            for i in slot_id
        ]
        return total_number_of_rows

    def _get_number_of_columns(self) -> List[int]:
        """
        Returns the total number of columns of the installed cards.
        """
        slot_id = self._get_slot_ids()
        total_number_of_columns = [int(float(self.ask(
            f'slot[{i}].columns.matrix')))
            for i in slot_id
        ]
        return total_number_of_columns

    def _get_rows(self) -> List[List[str]]:
        """
        Returns the elements of each row.
        """
        total_number_of_rows = self._get_number_of_rows()
        row_list = []
        for item in total_number_of_rows:
            rows_in_each_slot = [str(i) for i in range(1, item + 1)]
            row_list.append(rows_in_each_slot)
        return row_list

    def _get_columns(self) -> List[List[str]]:
        """
        Returns the elements of each column.
        """
        total_number_of_columns = self._get_number_of_columns()
        column_list = []
        for item in total_number_of_columns:
            columns_in_each_slot = []
            for i in range(1, item + 1):
                if i < 10:
                    columns_in_each_slot.append('0' + str(i))
                else:
                    columns_in_each_slot.append(str(i))
            column_list.append(columns_in_each_slot)
        return column_list

    def _get_channel_ranges(self) -> List[str]:
        """
        A helper function that gets two channel names from the available
        channels list and join them via a colon to define a channel range.
        """
        range_list = []
        for i in self._get_slot_ids():
            channel = self.get_channels_by_slot(int(i))
            for element in itertools.combinations(channel, 2):
                range_list.append(':'.join(element))
        return range_list

    def get_channels(self) -> List[str]:
        """
        This function returns the name of the matrix channels.
        User can call this function to see the names of the available
        channels, in case he/she is not familiar with the naming convention.
        However, note that, this is a standalone helper function and
        the usage of channel attributes of the instrument driver does
        not depend on the functionality of this method.
        """
        slot_id = self._get_slot_ids()
        row_list = self._get_rows()
        column_list = self._get_columns()
        matrix_channels = []
        for i, slot in enumerate(slot_id):
            for element in itertools.product(slot, row_list[i], column_list[i]):
                matrix_channels.append(''.join(element))
        return matrix_channels

    def get_channels_by_slot(self, slot_no: int) -> List[str]:
        """
        Returns the channel names of a given slot.

        Args:
            slot_no: An integer value specifying the slot number.
        """
        slot_id = self._get_slot_ids()
        if str(slot_no) not in slot_id:
            raise UnknownOrEmptySlot("Please provide a valid slot identifier. "
                                     f'Available slots are {slot_id}.')
        row_list = self._get_rows()
        column_list = self._get_columns()
        matrix_channels_by_slot = []
        for element in itertools.product(str(slot_no), row_list[0],
                                         column_list[0]):
            matrix_channels_by_slot.append(''.join(element))
        return matrix_channels_by_slot

    def get_analog_backplane_specifiers(self) -> List[str]:
        """
        Returns a list of comma separated strings representing available analog
        backplane relays. This function should not be mixed with the
        `get_backplane` method. The latter returns backplane relays which are
        associated with a channel by using `set_backplane` method.
        """
        backplane_common_number = '9'
        backplane_relay_common_numbers = ['11', '12', '13', '14', '15', '16']
        slot_id = self._get_slot_ids()
        analog_backplane_relays = []
        for element in itertools.product(slot_id, backplane_common_number,
                                         backplane_relay_common_numbers):
            analog_backplane_relays.append(''.join(element))
        return analog_backplane_relays

    def _connect_or_disconnect_row_to_columns(self, action: str,
                                              slot_id: int, row_id: int,
                                              columns: List[int]) -> List[str]:
        """
        A private function that connects or (disconnects) given columns
        to (from) a row of a slot and opens (closes) the formed channels.
        """
        if action not in ["connect", "disconnect"]:
            raise ValueError("The action should be identified as either "
                             "'connect' or 'disconnect'.")
        slots = self._get_slot_ids()
        slot = str(slot_id)
        if slot not in slots:
            raise UnknownOrEmptySlot("Please provide a valid slot identifier. "
                                     f'Available slots are {slots}.')
        row = str(row_id)
        columns_list = []
        for i in columns:
            if i < 10:
                columns_list.append('0' + str(i))
            else:
                columns_list.append(str(i))
        channels_to_connect_or_disconnect = []
        for element in itertools.product(slot, row, columns_list):
            channels_to_connect_or_disconnect.append(''.join(element))
        for channel in channels_to_connect_or_disconnect:
            if action == "connect":
                self.open_channel(channel)
            else:
                self.close_channel(channel)
        return channels_to_connect_or_disconnect

    def _connect_or_disconnect_column_to_rows(self, action: str,
                                              slot_id: int, column_id: int,
                                              rows: List[int]) -> List[str]:
        """
        A private function that connects (disconnects) given rows
        to (from) a column of a slot and opens (closes) the formed channels.
        """
        if action not in ["connect", "disconnect"]:
            raise ValueError("The action should be identified as either "
                             "'connect' or 'disconnect'.")
        slots = self._get_slot_ids()
        slot = str(slot_id)
        if slot not in slots:
            raise UnknownOrEmptySlot("Please provide a valid slot identifier. "
                                     f'Available slots are {slots}.')
        column = []
        if column_id < 10:
            column.append('0' + str(column_id))
        else:
            column.append(str(column_id))
        rows_list = [str(x) for x in rows]
        channels_to_connect_or_disconnect = []
        for element in itertools.product(slot, rows_list, column):
            channels_to_connect_or_disconnect.append(''.join(element))
        for channel in channels_to_connect_or_disconnect:
            if action == "connect":
                self.open_channel(channel)
            else:
                self.close_channel(channel)
        return channels_to_connect_or_disconnect

    def connect_row_to_columns(self, slot_id: int,
                               row_id: int, columns: List[int]) -> List[str]:
        """
        A convenient function that connects given columns to a row of a
        slot and opens the formed channels.

        Args:
            slot_id: The specifier for the slot from which the row and columns
                will be selected.
            row_id: The specifier for the row to which the provided columns
                will be connected.
            columns: The specifiers of the columns will be connected to the
                provided row.
        """
        return self._connect_or_disconnect_row_to_columns("connect", slot_id,
                                                          row_id, columns)

    def disconnect_row_from_columns(self, slot_id: int,
                                    row_id: int,
                                    columns: List[int]) -> List[str]:
        """
        A convenient function that disconnects given columns to a row of a
        slot and closes the formed channels.

        Args:
            slot_id: The specifier for the slot from which the row and columns
                will be selected.
            row_id: The specifier for the row to which the provided columns
                will be disconnected.
            columns: The specifiers of the columns will be disconnected from the
                provided row.
        """
        return self._connect_or_disconnect_row_to_columns("disconnect", slot_id,
                                                          row_id, columns)

    def connect_column_to_rows(self, slot_id: int,
                               column_id: int, rows: List[int]) -> List[str]:
        """
        A convenient function that connects given rows to a column of a
        slot and opens the formed channels.

        Args:
            slot_id: The specifier for the slot from which the row and columns
                will be selected.
            column_id: The specifier for the column to which the provided rows
                will be connected.
            rows: The specifiers of the rows will be connected to the
                provided column.
        """
        return self._connect_or_disconnect_column_to_rows("connect", slot_id,
                                                          column_id, rows)

    def disconnect_column_from_rows(self, slot_id: int,
                                    column_id: int,
                                    rows: List[int]) -> List[str]:
        """
        A convenient function that disconnects given rows to a column of a
        slot and closes the formed channels.

        Args:
            slot_id: The specifier for the slot from which the row and columns
                will be selected.
            column_id: The specifier for the column to which the provided rows
                will be disconnected.
            rows: The specifiers of the rows will be disconnected from the
                provided column.
        """
        return self._connect_or_disconnect_column_to_rows("disconnect", slot_id,
                                                          column_id, rows)

    def get_idn(self) -> Dict[str, Optional[str]]:
        """
        Overwrites the generic QCoDeS get IDN method. Returns
        a dictionary including the vendor, model, serial number and
        firmware version of the instrument.
        """
        idnstr = self.ask_raw('*IDN?')
        vendor, model, serial, firmware = map(str.strip, idnstr.split(','))
        model = model[6:]

        idn: Dict[str, Optional[str]] = {'vendor': vendor, 'model': model,
                                         'serial': serial, 'firmware': firmware}
        return idn

    def get_switch_cards(self) -> Tuple[Dict[str, str], ...]:
        """
        Returns a list of dictionaries listing the properties of the installed
        switch cards including the slot number tha it is installed, model,
        firmware version and serial number.
        """
        switch_cards: List[Dict[str, str]] = []
        for i in range(1, 7):
            scard = self.ask(f'slot[{i}].idn')
            if scard != 'Empty Slot':
                model, mtype, firmware, serial = map(str.strip,
                                                     scard.split(','))
                sdict = {'slot_no': str(i), 'model': model, 'mtype': mtype,
                         'firmware': firmware, 'serial': serial}
                switch_cards.append(sdict)
        return tuple(switch_cards)

    def get_available_memory(self) -> Dict[str, Optional[str]]:
        """
        Returns the amount of memory that is currently available for
        storing scripts, configurations and channel patterns.
        """
        memstring = self.ask('memory.available()')
        system_memory, script_memory, \
            pattern_memory, config_memory = map(str.strip, memstring.split(','))

        memory_available: Dict[str, Optional[str]] = {
            'System Memory  (%)': system_memory,
            'Script Memory  (%)': script_memory,
            'Pattern Memory (%)': pattern_memory,
            'Config Memory  (%)': config_memory
        }
        return memory_available

    def get_interlock_state(self) -> Tuple[Dict[str, str], ...]:
        """
        A function that collects the interlock status of the installed cards.
        The channel relays can continue to operate even if the interlock
        in the corresponding slot is disengaged, one cannot perform
        measurements through the switching card, as the analog backplanes
        cannot be energized.
        """
        slot_id = self._get_slot_ids()
        interlock_status = {0: 'Interlock is disengaged',
                            1: 'Interlock is engaged'}
        states: List[Dict[str, str]] = []
        for i in slot_id:
            state = self.get_interlock_state_by_slot(i)
            states.append({"slot_no": i, "state": interlock_status[state]})
        return tuple(states)

    def get_interlock_state_by_slot(self, slot: Union[str, int]) -> int:
        return int(float(self.ask(f"slot[{int(slot)}].interlock.state")))

    def get_ip_address(self) -> str:
        """
        Returns the current IP address of the instrument.
        """
        return self.ask('lan.status.ipaddress')

    def reset_local_network(self) -> None:
        """
        Resets the local network (LAN).
        """
        self.write('lan.reset()')

    def save_setup(self, val: Optional[str] = None) -> None:
        """
        Saves the present setup.

        Args:
            val: An optional string representing the path and the file name
                to which the setup shall be saved on a USB flash drive. If not
                provided, the setup will be saved to the nonvolatile memory
                of the instrument, any previous saves will be overwritten.
        """
        if val is not None:
            self.write(f"setup.save('{val}')")
        else:
            self.write(f'setup.save()')

    def load_setup(self, val: Union[int, str]) -> None:
        """
        Loads the settings from a saved setup.

        Args:
            val: An integer or a string that specifies the location of saved
                setup. If it is `0`, factory defaults load. If it is `1`,
                the saved setup from the nonvolatile memory is recalled.
                Otherwise, a string specifying the relative path to the saved
                setup on a USB drive should be passed in.
        """
        self.write(f"setup.recall('{val}')")

    def _validator(self, val: str) -> bool:
        """
        Instrument specific validator. As the number of validation points
        are around 15k, to avoid QCoDeS parameter validation to print them all,
        we shall raise a custom exception.
        """
        ch = self.get_channels()
        ch_range = self._get_channel_ranges()
        slots = ['allslots', *self._get_slot_names()]
        backplanes = self.get_analog_backplane_specifiers()
        specifier = val.split(',')
        for element in specifier:
            if element not in (*ch, *ch_range, *slots, *backplanes):
                return False
        return True

    def connect_message(self, idn_param: str = 'IDN',
                        begin_time: Optional[float] = None) -> None:
        """
        Overwrites the generic QCoDeS instrument connect message.
        Here, additionally, we provide information about
        which slots of the system switch is occupied with what
        kind of matrix, as well.
        """
        idn = self.get_idn()
        cards = self.get_switch_cards()
        states = self.get_interlock_state()

        con_msg = ('Connected to: {vendor} {model} SYSTEM SWITCH '
                   '(serial:{serial}, firmware:{firmware})'.format(**idn))
        print(con_msg)
        self.log.info(f"Connected to instrument: {idn}")

        for _, item in enumerate(cards):
            card_info = ('Slot {slot_no}- Model:{model}, Matrix Type:{mtype}, '
                         'Firmware:{firmware}, Serial:{serial}'.format(**item))
            print(card_info)
            self.log.info(f'Switch Cards: {item}')

        for _, item in enumerate(states):
            state_info = ('{state} in Slot {slot_no}.'.format(**item))
            print(state_info)

    def ask(self, cmd: str) -> str:
        """
        Override of normal ask. This is important, since queries to the
        instrument must be wrapped in 'print()'
        """
        return super().ask(f'print({cmd:s})')



"""
Slack bot is used to send information about qcodes via Slack IMs.
Some default commands are provided, and custom commands/tasks can be
attached (see below).

To setup the Slack bot, a bot first has to be created via Slack
by clicking 'Create New App' on https://api.slack.com/apps.
Once created, the bot will have a name and unique token.
These and other settings have to be saved in a config dict (see init( or
Parameters) in :class:`Slack`).

The App containing your bot needs to have the following bot token scopes to
perform all methods successfully:
- channels:history
- channels:read
- chat:write
- files:write
- users:read
These can be set after clicking OAuth & Permissions in the left menubar after
selecting your bot at https://api.slack.com/apps (or during creation).

Communication with the Slack bot is performed via instant messaging.
When an IM is sent to the Slack bot, it will be processed during the next
`update()` call (provided the username is registered in the config).
Standard commands provided to the Slack bot are:

- plot: Upload latest qcodes plot.
- msmt/measurement: Print information about latest measurement.
- notify finished: Send message once measurement is finished.

Custom commands can be added as (cmd, func) key-value pairs to
`self.commands`. When `cmd` is sent to the bot, `func` is evaluated.

Custom tasks can be added as well. These are functions that are performed
every time an update is called. The function must return a boolean that
indicates if the task should be removed from the list of tasks.
A custom task can be added as a (cmd, func) key-value pair  to
`self.task_commands`.
They can then be called through Slack IM via:

``notify/task {cmd} *args:`` register task with name `cmd` that is
performed every time `update()` is called.
"""

import inspect
import logging
import os
import tempfile
import threading
import traceback
import warnings
from functools import partial
from time import sleep

from requests.exceptions import ConnectTimeout, HTTPError, ReadTimeout
from requests.packages.urllib3.exceptions import ReadTimeoutError
from slack_sdk import WebClient

from qcodes import config as qc_config
from qcodes.instrument.parameter import _BaseParameter
from qcodes.loops import active_data_set, active_loop
from qcodes.plots.base import BasePlot


class SlackTimeoutWarning(UserWarning):
    pass


def convert_command(text):
    def try_convert_str(string):
        try:
            val = int(string)
            return val
        except ValueError:
            pass
        try:
            val = float(string)
            return val
        except ValueError:
            pass

        return string

    # Format text to lowercase, and remove trailing whitespaces
    text = text.lower().rstrip(' ')
    command, *args_str = text.split(' ')

    # Convert string args to floats/kwargs
    args = []
    kwargs = {}
    for arg in args_str:
        if '=' in arg:
            # arg is a kwarg
            key, val = arg.split('=')
            # Try to convert into a float
            val = try_convert_str(val)
            kwargs[key] = val
        else:
            # arg is not a kwarg
            # Try to convert into a float
            val = try_convert_str(arg)
            args.append(val)
    return command, args, kwargs


class Slack(threading.Thread):

    def __init__(self, interval=3, config=None, auto_start=True, **commands):
        """
        Initializes Slack bot, including auto-updating widget if in notebook
        and using multiprocessing.

        Args:
            interval (int): Update interval for widget (must be over 1s).
            config (Optional[dict]): Config dict
                If not given, uses qc.config['user']['slack']
                The config dict must contain the following keys:

                - 'bot_name': Name of the bot
                - 'bot_token': Token from bot (obtained from slack website)
                - 'names': Usernames to periodically check for IM messages

            auto_start (bool): Defaults to True.

        """
        if config is not None:
            self.config = config
        else:
            self.config = qc_config.user.slack

        self.slack = WebClient(token=self.config['token'])
        self.users = self.get_users(self.config['names'])
        self.get_im_ids(self.users)

        self.commands = {'plot': self.upload_latest_plot,
                         'msmt': self.print_measurement_information,
                         'measurement': self.print_measurement_information,
                         'notify': self.add_task,
                         'help': self.help_message,
                         'task': self.add_task,
                         **commands}
        self.task_commands = {'finished': self.check_msmt_finished}

        self.interval = interval
        self.tasks = []

        # Flag that exits loop when set to True (called via self.exit())
        self._exit = False

        # Flag that enables actions to be performed in the event loop
        # Enabled via self.start(), disabled via self.stop()
        self._is_active = False

        # Call Thread init
        super().__init__()

        if auto_start:
            self.start()

    def start(self):
        self._is_active = True
        try:
            # Start thread, can only be called once
            super().start()
        except RuntimeError:
            # Thread already started, ignoring
            pass

    def run(self):
        """
        Thread event loop that periodically checks for updates.
        Can be stopped via  :meth:`stop` , after which the Thread is stopped.
        Returns:
            None.
        """
        while not self._exit:
            # Continue event loop
            if self._is_active:
                # check for updates
                self.update()
            sleep(self.interval)

    def stop(self):
        """
        Stop checking for updates. Can be started again via :meth:`start`.
        Returns:
            None.
        """
        self._is_active = False

    def exit(self):
        """
        Exit event loop, stop Thread.
        Returns:
            None
        """
        self._stop = True

    def user_from_id(self, user_id):
        """
        Retrieve user from user id.
        Args:
            user_id: Id from which to retrieve user information.

        Returns:
            dict: User information.
        """
        return self.slack.users_info(user=user_id)['user']

    def get_users(self, usernames):
        """
        Extracts user information for users.
        Args:
            usernames: Slack usernames of users.

        Returns:
            dict: {username: user}
        """
        users = {}
        response = self.slack.users_list()
        for member in response['members']:
            if member['name'] in usernames:
                users[member['name']] = member
        if len(users) != len(usernames):
            remaining_names = [name for name in usernames if name not in users]
            raise RuntimeError(
                f'Could not find names {remaining_names}')
        return users

    def get_im_ids(self, users):
        """
        Adds IM ids of users to users dict.
        Also adds `last_ts` to the latest IM message
        Args:
            users (dict): {username: user}

        Returns:
            None.
        """
        response = self.slack.conversations_list(types='im')
        user_ids = {username: user['id'] for username, user in users.items()}
        im_ids = {chan['user']: chan['id'] for chan in response['channels']}
        for username, user_id in user_ids.items():
            if user_id in im_ids.keys():
                users[username]['im_id'] = im_ids[user_id]
                # update last ts
                messages = self.get_im_messages(username=username, limit=1)
                if messages:
                    users[username]['last_ts'] = float(messages[0]['ts'])
                else:
                    users[username]['last_ts'] = None

    def get_im_messages(self, username, **kwargs):
        """
        Retrieves IM messages from username.
        Args:
            username: Name of user.
            **kwargs: Additional kwargs for retrieving IM messages.

        Returns:
            List of IM messages.
        """
        # provide backward compatibility with 'count' keyword. It still works,
        # but is undocumented. 'count' likely does the same as 'limit', but
        # 'limit' takes precedence
        if 'limit' not in kwargs.keys():
            kwargs['limit'] = kwargs.pop('count', None)

        channel = self.users[username].get('im_id', None)
        if channel is None:
            return []
        else:
            response = self.slack.conversations_history(channel=channel,
                                                        **kwargs)
            return response['messages']

    def get_new_im_messages(self):
        """
        Retrieves new IM messages for each user in self.users.
        Updates user['last_ts'] to ts of newest message.
        Returns:
            im_messages (Dict): {username: [messages list]} newer than last_ts.
        """
        im_messages = {}
        for username, user in self.users.items():
            last_ts = user.get('last_ts', None)
            new_messages = self.get_im_messages(username=username,
                                                oldest=last_ts)
            # Kwarg 'oldest' sometimes also returns message with ts==last_ts
            new_messages = [m for m in new_messages if
                            float(m['ts']) != last_ts]
            im_messages[username] = new_messages
            if new_messages:
                self.users[username]['last_ts'] = float(new_messages[0]['ts'])
        return im_messages

    def update(self):
        """
        Performs tasks, and checks for new messages.
        Periodically called from widget update.
        Returns:
            None.
        """
        new_tasks = []
        for task in self.tasks:
            task_finished = task()
            if not task_finished:
                new_tasks.append(task)
        self.tasks = new_tasks

        new_messages = {}
        try:
            new_messages = self.get_new_im_messages()
        except (ReadTimeout, HTTPError, ConnectTimeout, ReadTimeoutError) as e:
            # catch any timeouts caused by network delays
            warnings.warn('error retrieving slack messages',
                          SlackTimeoutWarning)
            logging.info(e)
        self.handle_messages(new_messages)

    def help_message(self):
        """Return simple help message"""
        cc = ", ".join("`" + str(k) + "`" for k in self.commands.keys())
        return "\nAvailable commands: %s" % cc

    def handle_messages(self, messages):
        """
        Performs commands depending on messages.
        This includes adding tasks to be performed during each update.
        """
        for user, user_messages in messages.items():
            for message in user_messages:
                if message.get('user', None) != self.users[user]['id']:
                    # Filter out bot messages
                    continue
                channel = self.users[user]['im_id']
                # Extract command (first word) and possible args
                command, args, kwargs = convert_command(message['text'])
                if command in self.commands:
                    msg = f'Executing {command}'
                    if args:
                        msg += f' {args}'
                    if kwargs:
                        msg += f' {kwargs}'
                    self.slack.chat_postMessage(text=msg, channel=channel)

                    func = self.commands[command]
                    try:
                        if isinstance(func, _BaseParameter):
                            results = func(*args, **kwargs)
                        else:
                            # Only add channel and Slack if they are explicit
                            # kwargs
                            func_sig = inspect.signature(func)
                            if 'channel' in func_sig.parameters:
                                kwargs['channel'] = channel
                            if 'slack' in func_sig.parameters:
                                kwargs['slack'] = self
                            results = func(*args, **kwargs)

                        if results is not None:
                            self.slack.chat_postMessage(
                                text=f'Results: {results}',
                                channel=channel)

                    except Exception:
                        self.slack.chat_postMessage(
                            text=f'Error: {traceback.format_exc()}',
                            channel=channel)
                else:
                    self.slack.chat_postMessage(
                        text='Command {} not understood. Try `help`'.format(
                            command),
                        channel=channel)

    def add_task(self, command, *args, channel, **kwargs):
        """
        Add a task to self.tasks, which will be executed during each update
        Args:
            command: Task command.
            *args: Additional args for command.
            channel: Slack channel (can also be IM channel).
            **kwargs: Additional kwargs for particular.

        Returns:
            None.
        """
        if command in self.task_commands:
            self.slack.chat_postMessage(
                text=f'Added task "{command}"',
                channel=channel)
            func = self.task_commands[command]
            self.tasks.append(partial(func, *args, channel=channel, **kwargs))
        else:
            self.slack.chat_postMessage(
                text=f'Task command {command} not understood',
                channel=channel)

    def upload_latest_plot(self, channel, **kwargs):
        """
        Uploads latest plot (if any) to slack channel.
        The latest plot is retrieved from
        :class:`qcodes.plots.base.BasePlot`, which is updated
        every time a new qcodes plot is instantiated.
        Args:
            channel: Slack channel (can also be IM channel).
            **kwargs: Not used.

        Returns:
            None.
        """
        # Create temporary filename
        temp_filename = tempfile.mktemp(suffix='.jpg')
        # Retrieve latest plot
        latest_plot = BasePlot.latest_plot
        if latest_plot is not None:
            # Saves latest plot to filename
            latest_plot.save(filename=temp_filename)
            # Upload plot to slack
            self.slack.files_upload(file=temp_filename, channels=channel)
            os.remove(temp_filename)
        else:
            self.slack.chat_postMessage(text='No latest plot',
                                        channel=channel)

    def print_measurement_information(self, channel, **kwargs):
        """
        Prints information about the current measurement.
        Information printed is percentage complete, and dataset representation.
        Dataset is retrieved from DataSet.latest_dataset, which updates itself
        every time a new dataset is created
        Args:
            channel: Slack channel (can also be IM channel).
            **kwargs: Not used.

        Returns:
            None.
        """
        dataset = active_data_set()
        if dataset is not None:
            self.slack.chat_postMessage(
                text='Measurement is {:.0f}% complete'.format(
                    100 * dataset.fraction_complete()),
                channel=channel)
            self.slack.chat_postMessage(
                text=repr(dataset), channel=channel)
        else:
            self.slack.chat_postMessage(
                text='No latest dataset found',
                channel=channel)

    def check_msmt_finished(self, channel, **kwargs):
        """
        Checks if the latest measurement is completed.
        Args:
            channel: Slack channel (can also be IM channel).
            **kwargs: Not used.

        Returns:
            bool: True if measurement is finished, False otherwise.
        """
        if active_loop() is None:
            self.slack.chat_postMessage(
                text='Measurement complete',
                channel=channel)
            return True
        else:
            return False



import re

import pytest

import qcodes.instrument_drivers.Keysight.keysightb1500.constants as c
from qcodes.instrument_drivers.Keysight.keysightb1500.message_builder import \
    MessageBuilder


@pytest.fixture
def mb():
    yield MessageBuilder()


def test_as_csv():
    from qcodes.instrument_drivers.Keysight.keysightb1500.message_builder \
        import as_csv

    assert '1' == as_csv([1])

    assert '1,2,3' == as_csv([1, 2, 3])

    assert '1,2' == as_csv([c.ChNr.SLOT_01_CH1, c.ChNr.SLOT_02_CH1])


def _skip():
    pytest.skip("not implemented yet")


def test_cmd(mb):
    cmd = mb.aad(1, 0).ach(1, 5).ab()
    assert 'AAD 1,0;ACH 1,5;AB' == cmd.message


def test_raise_error_on_appending_command_after_final_command(mb):
    with pytest.raises(ValueError):
        mb.aad(1, 0).ab().ach(1, 5)


def test_exception_on_too_long_message(mb):
    length = 0
    while length < 250:
        mb.os()
        length += 3  # "OS;" <- three characters per iteration

    match = re.escape(
        f"Command is too long ({len(str(mb._msg))}>256-termchars) "
        f"and will overflow input buffer of instrument. "
        f"(Consider using the ST/END/DO/RU commands for very long "
        f"programs.)")
    with pytest.raises(Exception, match=match):
        _ = mb.message


def test_clear_message_queue(mb):
    mb.aad(1, 0)
    assert mb.message == 'AAD 1,0'

    mb.clear_message_queue()

    assert mb.message == ''


def test_aad(mb):
    assert 'AAD 1,0' == \
           mb.aad(c.ChNr.SLOT_01_CH1,
                  c.AAD.Type.HIGH_SPEED).message

    mb.clear_message_queue()

    assert 'AAD 1,1' == \
           mb.aad(c.ChNr.SLOT_01_CH1,
                  adc_type=c.AAD.Type.HIGH_RESOLUTION).message


def test_ab(mb):
    assert 'AB' == mb.ab().message


def test_ach(mb):
    assert 'ACH 1,5' == mb.ach(c.ChNr.SLOT_01_CH1,
                               c.ChNr.SLOT_05_CH1).message

    mb.clear_message_queue()

    assert 'ACH 1' == mb.ach(c.ChNr.SLOT_01_CH1).message

    mb.clear_message_queue()

    assert 'ACH' == mb.ach().message


def test_act(mb):
    assert 'ACT 0,1' == mb.act(c.ACT.Mode.AUTO, 1).message

    mb.clear_message_queue()

    assert 'ACT 2' == mb.act(c.ACT.Mode.PLC).message


def test_acv(mb):
    assert 'ACV 7,0.01' == \
           mb.acv(c.ChNr.SLOT_07_CH1, 0.01).message


def test_adj(mb):
    assert 'ADJ 9,1' == mb.adj(c.ChNr.SLOT_09_CH1,
                               c.ADJ.Mode.MANUAL).message


def test_adj_query(mb):
    assert 'ADJ? 1' == mb.adj_query(c.ChNr.SLOT_01_CH1).message

    mb.clear_message_queue()

    assert 'ADJ? 1,1' == mb.adj_query(c.ChNr.SLOT_01_CH1,
                                      c.ADJQuery.Mode.MEASURE).message


def test_ait(mb):
    assert 'AIT 2,3,0.001' == mb.ait(c.AIT.Type.HIGH_SPEED_PULSED,
                                     c.AIT.Mode.MEAS_TIME_MODE,
                                     0.001).message
    mb.clear_message_queue()
    assert 'AIT 2,3' == mb.ait(c.AIT.Type.HIGH_SPEED_PULSED,
                               c.AIT.Mode.MEAS_TIME_MODE).message


def test_aitm(mb):
    assert 'AITM 0' == mb.aitm(c.APIVersion.B1500).message


def test_aitm_query(mb: MessageBuilder):
    assert 'AITM?' == mb.aitm_query().message


def test_als(mb):
    with pytest.raises(NotImplementedError):
        mb.als(c.ChNr.SLOT_01_CH1, 1, b'a')


def test_als_query(mb):
    assert 'ALS? 1' == mb.als_query(c.ChNr.SLOT_01_CH1).message


def test_alw(mb):
    with pytest.raises(NotImplementedError):
        mb.alw(c.ChNr.SLOT_01_CH1, 1, b'a')


def test_alw_query(mb):
    assert 'ALW? 1' == mb.alw_query(c.ChNr.SLOT_01_CH1).message


def test_av(mb):
    assert 'AV 10' == mb.av(10).message
    mb.clear_message_queue()
    assert 'AV -50' == mb.av(-50).message
    mb.clear_message_queue()
    assert 'AV 100,1' == mb.av(100, c.AV.Mode.MANUAL).message


def test_az(mb):
    assert 'AZ 0' == mb.az(False).message
    mb.clear_message_queue()
    assert 'AZ 1' == mb.az(True).message


def test_bc(mb):
    assert 'BC' == mb.bc().message


def test_bdm(mb: MessageBuilder):
    assert 'BDM 0,1' == mb.bdm(c.BDM.Interval.SHORT,
                               c.BDM.Mode.CURRENT).message


def test_bdt(mb: MessageBuilder):
    assert 'BDT 0.1,0.001' == mb.bdt(hold=0.1, delay=1e-3).message


def test_bdv(mb):
    assert 'BDV 1,0,0,100,0.01' == mb.bdv(chnum=c.ChNr.SLOT_01_CH1,
                                          v_range=c.VOutputRange.AUTO,
                                          start=0, stop=100,
                                          i_comp=0.01).message


def test_bgi(mb):
    assert 'BGI 1,0,1e-08,14,1e-06' == \
           mb.bgi(chnum=c.ChNr.SLOT_01_CH1,
                  searchmode=c.BinarySearchMode.LIMIT,
                  stop_condition=1e-8,
                  i_range=14,
                  target=1e-6).message


def test_bgv(mb):
    assert 'BGV 1,0,0.1,12,5' == mb.bgv(1, 0, 0.1, 12, 5).message


def test_bsi(mb):
    assert 'BSI 1,0,1e-12,1e-06,10' == mb.bsi(1, 0, 1e-12, 1e-6,
                                              10).message


def test_bsm(mb):
    assert 'BSM 1,2,3' == mb.bsm(1, 2, 3).message


def test_bssi(mb):
    assert 'BSSI 1,0,1e-06,10' == mb.bssi(1, 0, 1e-6, 10).message


def test_bssv(mb):
    assert 'BSSV 1,0,5,1e-06' == mb.bssv(1, 0, 5, 1e-6).message


def test_bst(mb):
    assert 'BST 5,0.1' == mb.bst(5, 0.1).message


def test_bsv(mb):
    assert 'BSV 1,0,0,20,1e-06' == mb.bsv(1, 0, 0, 20, 1e-6).message


def test_bsvm(mb):
    assert 'BSVM 1' == mb.bsvm(1).message


def test_ca(mb):
    assert 'CA' == mb.ca().message


def test_cal_query(mb):
    assert '*CAL?' == mb.cal_query().message


def test_cl(mb):
    assert 'CL' == mb.cl().message
    mb.clear_message_queue()
    assert 'CL 1,2,3,5' == mb.cl([1, 2, 3, 5]).message


def test_clcorr(mb):
    assert 'CLCORR 9,1' == mb.clcorr(9, 1).message


def test_cm(mb):
    assert 'CM 0' == mb.cm(False).message


def test_cmm(mb):
    assert 'CMM 1,2' == mb.cmm(1, 2).message


def test_cn(mb):
    assert 'CN' == mb.cn().message
    mb.clear_message_queue()
    assert 'CN 1,2,3,5' == mb.cn([1, 2, 3, 5]).message


def test_cnx(mb):
    assert 'CNX' == mb.cnx().message
    mb.clear_message_queue()
    assert 'CNX 1,2,3,5' == mb.cnx([1, 2, 3, 5]).message


def test_corr_query(mb):
    assert 'CORR? 9,3' == mb.corr_query(9, 3).message


def test_corrdt(mb):
    assert 'CORRDT 9,3000000,0,0,0,0,0,0' == mb.corrdt(9, 3000000, 0,
                                                       0, 0, 0, 0,
                                                       0).message


def test_corrdt_query(mb):
    assert 'CORRDT? 9,1' == mb.corrdt_query(9, 1).message


def test_corrl(mb):
    assert 'CORRL 9,3000000' == mb.corrl(9, 3000000).message


def test_corrl_query(mb):
    assert 'CORRL? 9' == mb.corrl_query(9).message
    mb.clear_message_queue()
    assert 'CORRL? 9,2' == mb.corrl_query(9, 2).message


def test_corrser_query(mb: MessageBuilder):
    assert 'CORRSER? 101,1,1e-07,1e-08,10' == \
           mb.corrser_query(101,
                            True,
                            1E-7,
                            1E-8,
                            10).message


def test_corrst(mb):
    assert 'CORRST 3,1,1' == mb.corrst(3, 1, 1).message


def test_corrst_query(mb):
    assert 'CORRST? 1,2' == mb.corrst_query(1, 2).message


def test_dcorr(mb):
    assert 'DCORR 3,1,100,0.1,0.2' == mb.dcorr(3, 1, 100, 0.1, 0.2).message


def test_dcorr_query(mb):
    assert 'DCORR? 3,1' == mb.dcorr_query(3, 1).message


def test_dcv(mb):
    assert 'DCV 3,2.5' == mb.dcv(3, 2.5).message


def test_di(mb):
    # assert '' == mb.().message
    _skip()


def test_diag_query(mb):
    # assert '' == mb.().message
    _skip()


def test_do(mb):
    # assert '' == mb.().message
    _skip()


def test_dsmplarm(mb):
    # assert '' == mb.().message
    _skip()


def test_dsmplflush(mb):
    # assert '' == mb.().message
    _skip()


def test_dsmplsetup(mb):
    # assert '' == mb.().message
    _skip()


def test_dv(mb):
    # assert '' == mb.().message
    _skip()


def test_dz(mb):
    # assert '' == mb.().message
    _skip()


def test_emg_query(mb):
    # assert '' == mb.().message
    _skip()


def test_end(mb):
    # assert '' == mb.().message
    _skip()


def test_erc(mb):
    # assert '' == mb.().message
    _skip()


def test_ercmaa(mb):
    # assert '' == mb.().message
    _skip()


def test_ercmaa_query(mb):
    # assert '' == mb.().message
    _skip()


def test_ercmagrd(mb):
    # assert '' == mb.().message
    _skip()


def test_ercmagrd_query(mb):
    # assert '' == mb.().message
    _skip()


def test_ercmaio(mb):
    # assert '' == mb.().message
    _skip()


def test_ercmaio_query(mb):
    # assert '' == mb.().message
    _skip()


def test_ercmapfgd(mb):
    # assert '' == mb.().message
    _skip()


def test_erhpa(mb):
    # assert '' == mb.().message
    _skip()


def test_erhpa_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erhpe(mb):
    # assert '' == mb.().message
    _skip()


def test_erhpe_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erhpl(mb):
    # assert '' == mb.().message
    _skip()


def test_erhpl_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erhpp(mb):
    # assert '' == mb.().message
    _skip()


def test_erhpp_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erhpqg(mb):
    # assert '' == mb.().message
    _skip()


def test_erhpqg_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erhpr(mb):
    # assert '' == mb.().message
    _skip()


def test_erhpr_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erhps(mb):
    # assert '' == mb.().message
    _skip()


def test_erhps_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erhvca(mb):
    # assert '' == mb.().message
    _skip()


def test_erhvca_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erhvctst_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erhvp(mb):
    # assert '' == mb.().message
    _skip()


def test_erhvp_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erhvpv(mb):
    # assert '' == mb.().message
    _skip()


def test_erhvs(mb):
    # assert '' == mb.().message
    _skip()


def test_erhvs_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erm(mb):
    # assert '' == mb.().message
    _skip()


def test_ermod(mb):
    # assert '' == mb.().message
    _skip()


def test_ermod_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfda(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfda_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfdp(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfdp_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfds(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfds_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfga(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfga_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfgp(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfgp_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfgr(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfgr_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfqg(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfqg_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erpftemp_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfuhca(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfuhca_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfuhccal_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfuhcmax_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erpfuhctst(mb):
    # assert '' == mb.().message
    _skip()


def test_err_query(mb):
    # assert '' == mb.().message
    _skip()


def test_errx_query(mb):
    assert 'ERRX?' == mb.errx_query().message
    mb.clear_message_queue()
    assert 'ERRX? 0' == mb.errx_query(mode=c.ERRX.Mode.CODE_AND_MESSAGE
                                      ).message


def test_ers_query(mb):
    # assert '' == mb.().message
    _skip()


def test_erssp(mb):
    # assert '' == mb.().message
    _skip()


def test_erssp_query(mb):
    # assert '' == mb.().message
    _skip()


def test_eruhva(mb):
    # assert '' == mb.().message
    _skip()


def test_eruhva_query(mb):
    # assert '' == mb.().message
    _skip()


def test_fc(mb):
    assert 'FC 3,1000' == mb.fc(3, 1000).message



def test_fl(mb):
    assert 'FL 1' == mb.fl(True).message
    mb.clear_message_queue()
    assert 'FL 0,1,3,5' == mb.fl(False, [1, 3, 5]).message
    mb.clear_message_queue()
    assert 'FL 0,1,2,3,4,5,6,7,8,9,10' == \
           mb.fl(False, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).message
    mb.clear_message_queue()
    with pytest.raises(ValueError):
        _ = mb.fl(False, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]).message


def test_fmt(mb):
    assert 'FMT 1' == \
           mb.fmt(c.FMT.Format.ASCII_12_DIGITS_WITH_HEADER_CRLF_EOI
                  ).message
    mb.clear_message_queue()
    assert 'FMT 2,1' == \
           mb.fmt(c.FMT.Format.ASCII_12_DIGITS_NO_HEADER_CRLF_EOI,
                  c.FMT.Mode.PRIMARY_SOURCE_OUTPUT_DATA).message


def test_hvsmuop(mb):
    _skip()


def test_hvsmuop_query(mb):
    assert 'HVSMUOP?' == \
           mb.hvsmuop_query().message


def test_idn_query(mb):
    assert 'IN' == mb.in_().message
    mb.clear_message_queue()
    assert 'IN 1,2,3,5,6' == mb.in_([1, 2, 3, 5, 6]).message


def test_imp(mb):
    assert 'IMP 10' == mb.imp(c.IMP.MeasurementMode.Z_THETA_RAD).message


def test_in_(mb):
    # assert '' == mb.().message
    _skip()


def test_intlkvth(mb):
    # assert '' == mb.().message
    _skip()


def test_intlkvth_query(mb):
    # assert '' == mb.().message
    _skip()


def test_lgi(mb):
    # assert '' == mb.().message
    _skip()


def test_lgv(mb):
    # assert '' == mb.().message
    _skip()


def test_lim(mb):
    # assert '' == mb.().message
    _skip()


def test_lim_query(mb):
    # assert '' == mb.().message
    _skip()


def test_lmn(mb):
    # assert '' == mb.().message
    _skip()


def test_lop_query(mb):
    # assert '' == mb.().message
    _skip()


def test_lrn_query(mb):
    # assert '' == mb.().message
    _skip()


def test_lsi(mb):
    # assert '' == mb.().message
    _skip()


def test_lsm(mb):
    # assert '' == mb.().message
    _skip()


def test_lssi(mb):
    # assert '' == mb.().message
    _skip()


def test_lssv(mb):
    # assert '' == mb.().message
    _skip()


def test_lst_query(mb):
    # assert '' == mb.().message
    _skip()


def test_lstm(mb):
    # assert '' == mb.().message
    _skip()


def test_lsv(mb):
    # assert '' == mb.().message
    _skip()


def test_lsvm(mb):
    # assert '' == mb.().message
    _skip()


def test_mcc(mb):
    # assert '' == mb.().message
    _skip()


def test_mcpnt(mb):
    # assert '' == mb.().message
    _skip()


def test_mcpnx(mb):
    # assert '' == mb.().message
    _skip()


def test_mcpt(mb):
    # assert '' == mb.().message
    _skip()


def test_mcpws(mb):
    # assert '' == mb.().message
    _skip()


def test_mcpwnx(mb):
    # assert '' == mb.().message
    _skip()


def test_mdcv(mb):
    # assert '' == mb.().message
    _skip()


def test_mi(mb):
    # assert '' == mb.().message
    _skip()


def test_ml(mb):
    # assert '' == mb.().message
    _skip()


def test_mm(mb):
    assert 'MM 1,1' == \
           mb.mm(mode=c.MM.Mode.SPOT,
                 channels=[c.ChNr.SLOT_01_CH1]).message
    mb.clear_message_queue()
    assert 'MM 2,1,3' == mb.mm(mode=c.MM.Mode.STAIRCASE_SWEEP,
                               channels=[c.ChNr.SLOT_01_CH1,
                                         c.ChNr.SLOT_03_CH1]).message


def test_msc(mb):
    # assert '' == mb.().message
    _skip()


def test_msp(mb):
    # assert '' == mb.().message
    _skip()


def test_mt(mb):
    assert 'MT 0.0,0.42,32' == mb.mt(0.0, 0.42, 32).message
    mb.clear_message_queue()
    assert 'MT 0.0,0.42,32,0.12' == mb.mt(0.0, 0.42, 32, 0.12).message


def test_mtdcv(mb):
    # assert '' == mb.().message
    _skip()


def test_mv(mb):
    # assert '' == mb.().message
    _skip()


def test_nub_query(mb):
    # assert '' == mb.().message
    _skip()


def test_odsw(mb):
    # assert '' == mb.().message
    _skip()


def test_odsw_query(mb):
    # assert '' == mb.().message
    _skip()


def test_opc_query(mb):
    # assert '' == mb.().message
    _skip()


def test_os(mb):
    # assert '' == mb.().message
    _skip()


def test_osx(mb):
    # assert '' == mb.().message
    _skip()


def test_pa(mb):
    # assert '' == mb.().message
    _skip()


def test_pad(mb):
    # assert '' == mb.().message
    _skip()


def test_pax(mb):
    # assert '' == mb.().message
    _skip()


def test_pch(mb):
    # assert '' == mb.().message
    _skip()


def test_pch_query(mb):
    # assert '' == mb.().message
    _skip()


def test_pdcv(mb):
    # assert '' == mb.().message
    _skip()


def test_pi(mb):
    # assert '' == mb.().message
    _skip()


def test_pt(mb):
    # assert '' == mb.().message
    _skip()


def test_ptdcv(mb):
    # assert '' == mb.().message
    _skip()


def test_pv(mb):
    # assert '' == mb.().message
    _skip()


def test_pwdcv(mb):
    # assert '' == mb.().message
    _skip()


def test_pwi(mb):
    # assert '' == mb.().message
    _skip()


def test_pwv(mb):
    # assert '' == mb.().message
    _skip()


def test_qsc(mb):
    # assert '' == mb.().message
    _skip()


def test_qsl(mb):
    # assert '' == mb.().message
    _skip()


def test_qsm(mb):
    # assert '' == mb.().message
    _skip()


def test_qso(mb):
    # assert '' == mb.().message
    _skip()


def test_qsr(mb):
    # assert '' == mb.().message
    _skip()


def test_qst(mb):
    # assert '' == mb.().message
    _skip()


def test_qsv(mb):
    # assert '' == mb.().message
    _skip()


def test_qsz(mb):
    # assert '' == mb.().message
    _skip()


def test_rc(mb):
    # assert '' == mb.().message
    _skip()


def test_rcv(mb):
    # assert '' == mb.().message
    _skip()


def test_ri(mb):
    # assert '' == mb.().message
    _skip()


def test_rm(mb):
    assert 'RM 1,2' == mb.rm(1, c.RM.Mode.AUTO_UP).message
    mb.clear_message_queue()
    assert 'RM 2,3,60' == mb.rm(2, c.RM.Mode.AUTO_UP_DOWN, 60).message
    mb.clear_message_queue()
    with pytest.raises(ValueError):
        mb.rm(c.ChNr.SLOT_01_CH1, c.RM.Mode.DEFAULT, 22)


def test_rst(mb):
    # assert '' == mb.().message
    _skip()


def test_ru(mb):
    # assert '' == mb.().message
    _skip()


def test_rv(mb):
    # assert '' == mb.().message
    _skip()


def test_rz(mb):
    # assert '' == mb.().message
    _skip()


def test_sal(mb):
    # assert '' == mb.().message
    _skip()


def test_sap(mb):
    # assert '' == mb.().message
    _skip()


def test_sar(mb):
    assert 'SAR 1,0' == mb.sar(1,
                               enable_picoamp_autoranging=True).message


def test_scr(mb):
    # assert '' == mb.().message
    _skip()


def test_ser(mb):
    # assert '' == mb.().message
    _skip()


def test_ser_query(mb):
    # assert '' == mb.().message
    _skip()


def test_sim(mb):
    # assert '' == mb.().message
    _skip()


def test_sim_query(mb):
    # assert '' == mb.().message
    _skip()


def test_sopc(mb):
    # assert '' == mb.().message
    _skip()


def test_sopc_query(mb):
    # assert '' == mb.().message
    _skip()


def test_sovc(mb):
    # assert '' == mb.().message
    _skip()


def test_sovc_query(mb):
    # assert '' == mb.().message
    _skip()


def test_spm(mb):
    # assert '' == mb.().message
    _skip()


def test_spm_query(mb):
    # assert '' == mb.().message
    _skip()


def test_spp(mb):
    # assert '' == mb.().message
    _skip()


def test_spper(mb):
    # assert '' == mb.().message
    _skip()


def test_spper_query(mb):
    # assert '' == mb.().message
    _skip()


def test_sprm(mb):
    # assert '' == mb.().message
    _skip()


def test_sprm_query(mb):
    # assert '' == mb.().message
    _skip()


def test_spst_query(mb):
    # assert '' == mb.().message
    _skip()


def test_spt(mb):
    # assert '' == mb.().message
    _skip()


def test_spt_query(mb):
    # assert '' == mb.().message
    _skip()


def test_spupd(mb):
    # assert '' == mb.().message
    _skip()


def test_spv(mb):
    # assert '' == mb.().message
    _skip()


def test_spv_query(mb):
    # assert '' == mb.().message
    _skip()


def test_sre(mb):
    # assert '' == mb.().message
    _skip()


def test_sre_query(mb):
    # assert '' == mb.().message
    _skip()


def test_srp(mb):
    # assert '' == mb.().message
    _skip()


def test_ssl(mb):
    assert 'SSL 9,0' == mb.ssl(9, enable_indicator_led=False).message


def test_ssp(mb):
    # assert '' == mb.().message
    _skip()


def test_ssr(mb):
    # assert '' == mb.().message
    _skip()


def test_st(mb):
    # assert '' == mb.().message
    _skip()


def test_stb_query(mb):
    # assert '' == mb.().message
    _skip()


def test_stgp(mb):
    # assert '' == mb.().message
    _skip()


def test_stgp_query(mb):
    # assert '' == mb.().message
    _skip()


def test_tacv(mb):
    # assert '' == mb.().message
    _skip()


def test_tc(mb):
    assert 'TC 3,0' == mb.tc(3, 0).message


def test_tdcv(mb):
    # assert '' == mb.().message
    _skip()


def test_tdi(mb):
    # assert '' == mb.().message
    _skip()


def test_tdv(mb):
    # assert '' == mb.().message
    _skip()


def test_tgmo(mb):
    # assert '' == mb.().message
    _skip()


def test_tgp(mb):
    # assert '' == mb.().message
    _skip()


def test_tgpc(mb):
    # assert '' == mb.().message
    _skip()


def test_tgsi(mb):
    # assert '' == mb.().message
    _skip()


def test_tgso(mb):
    # assert '' == mb.().message
    _skip()


def test_tgxo(mb):
    # assert '' == mb.().message
    _skip()


def test_ti(mb):
    assert 'TI 1' == mb.ti(chnum=1).message
    mb.clear_message_queue()
    assert 'TI 1,-14' == mb.ti(chnum=1, i_range=c.IMeasRange.FIX_1uA).message


def test_tiv(mb):
    # assert '' == mb.().message
    _skip()


def test_tm(mb):
    # assert '' == mb.().message
    _skip()


def test_tmacv(mb):
    # assert '' == mb.().message
    _skip()


def test_tmdcv(mb):
    # assert '' == mb.().message
    _skip()


def test_tsc(mb):
    # assert '' == mb.().message
    _skip()


def test_tsq(mb):
    # assert '' == mb.().message
    _skip()


def test_tsr(mb):
    assert 'TSR' == mb.tsr().message
    mb.clear_message_queue()
    assert 'TSR 1' == mb.tsr(c.ChNr.SLOT_01_CH1).message


def test_tst(mb):
    # assert '' == mb.().message
    _skip()


def test_ttc(mb):
    # assert '' == mb.().message
    _skip()


def test_tti(mb):
    # assert '' == mb.().message
    _skip()


def test_ttiv(mb):
    # assert '' == mb.().message
    _skip()


def test_ttv(mb):
    # assert '' == mb.().message
    _skip()


def test_tv(mb):
    # assert '' == mb.().message
    _skip()


def test_unt_query(mb):
    # assert '' == mb.().message
    _skip()


def test_var(mb):
    # assert '' == mb.().message
    _skip()


def test_var_query(mb):
    # assert '' == mb.().message
    _skip()


def test_wacv(mb):
    # assert '' == mb.().message
    _skip()


def test_wat(mb):
    # assert '' == mb.().message
    _skip()


def test_wdcv(mb):
    # assert '' == mb.().message
    _skip()


def test_wfc(mb):
    # assert '' == mb.().message
    _skip()


def test_wi(mb):
    # assert '' == mb.().message
    _skip()


def test_wm(mb):
    # assert '' == mb.().message
    _skip()


def test_wmacv(mb):
    # assert '' == mb.().message
    _skip()


def test_wmdcv(mb):
    # assert '' == mb.().message
    _skip()


def test_wmfc(mb):
    # assert '' == mb.().message
    _skip()


def test_wncc(mb):
    # assert '' == mb.().message
    _skip()


def test_wnu_query(mb):
    # assert '' == mb.().message
    _skip()


def test_wnx(mb):
    # assert '' == mb.().message
    _skip()


def test_ws(mb):
    # assert '' == mb.().message
    _skip()


def test_wsi(mb):
    # assert '' == mb.().message
    _skip()


def test_wsv(mb):
    # assert '' == mb.().message
    _skip()


def test_wt(mb):
    # assert '' == mb.().message
    _skip()


def test_wtacv(mb):
    # assert '' == mb.().message
    _skip()


def test_wtdcv(mb):
    # assert '' == mb.().message
    _skip()


def test_wtfc(mb):
    # assert '' == mb.().message
    _skip()


def test_wv(mb):
    # assert '' == mb.().message
    _skip()


def test_wz_query(mb):
    # assert '' == mb.().message
    _skip()


def test_xe(mb):
    assert 'XE' == mb.xe().message


def test_nplc_setting_for_high_speed_vs_high_resolution_mode(mb):
    msg = (mb
           .ait(adc_type=c.AIT.Type.HIGH_SPEED,
                mode=c.AIT.Mode.NPLC,
                coeff=3)
           .ait(adc_type=c.AIT.Type.HIGH_RESOLUTION,
                mode=c.AIT.Mode.NPLC,
                coeff=8)
           .message
           )
    assert msg == 'AIT 0,2,3;AIT 1,2,8'


def test_set_resolution_mode_for_each_smu(mb):
    msg = (mb
           .aad(chnum=1, adc_type=c.AAD.Type.HIGH_SPEED)
           .aad(chnum=2, adc_type=c.AAD.Type.HIGH_RESOLUTION)
           .message
           )
    assert msg == 'AAD 1,0;AAD 2,1'



from unittest import TestCase
from datetime import datetime

from qcodes.data.location import FormatLocation
from qcodes.instrument.parameter import Parameter
from qcodes.measure import Measure

from ..instrument_mocks import MultiGetter, MultiSetPointParam

import numpy as np
from numpy.testing import assert_array_equal

class TestMeasure(TestCase):
    def setUp(self):
        self.p1 = Parameter('P1', initial_value=1, get_cmd=None, set_cmd=None)

    def test_simple_scalar(self):
        data = Measure(self.p1).run_temp()

        self.assertEqual(data.single_set.tolist(), [0])
        self.assertEqual(data.P1.tolist(), [1])
        self.assertEqual(len(data.arrays), 2, data.arrays)

        self.assertNotIn('loop', data.metadata)

        meta = data.metadata['measurement']
        self.assertEqual(meta['__class__'], 'qcodes.measure.Measure')
        self.assertEqual(len(meta['actions']), 1)
        self.assertFalse(meta['use_threads'])

        ts_start = datetime.strptime(meta['ts_start'], '%Y-%m-%d %H:%M:%S')
        ts_end = datetime.strptime(meta['ts_end'], '%Y-%m-%d %H:%M:%S')
        self.assertGreaterEqual(ts_end, ts_start)

    def test_simple_array(self):
        data = Measure(MultiGetter(arr=(1.2, 3.4))).run_temp()

        self.assertEqual(data.index0_set.tolist(), [0, 1])
        self.assertEqual(data.arr.tolist(), [1.2, 3.4])
        self.assertEqual(len(data.arrays), 2, data.arrays)

    def test_array_and_scalar(self):
        self.p1.set(42)
        data = Measure(MultiGetter(arr=(5, 6)), self.p1).run_temp()

        self.assertEqual(data.single_set.tolist(), [0])
        self.assertEqual(data.P1.tolist(), [42])
        self.assertEqual(data.index0_set.tolist(), [0, 1])
        self.assertEqual(data.arr.tolist(), [5, 6])
        self.assertEqual(len(data.arrays), 4, data.arrays)


class TestMeasureMulitParameter(TestCase):
    def setUp(self):
        self.p1 = MultiSetPointParam()

    def test_metadata(self):
        loc_fmt = 'data/{date}/#{counter}_{name}_{date}_{time}'
        rcd = {'name': 'test_metadata'}
        param_name_1 = "multi_setpoint_param_this"
        param_name_2 = "multi_setpoint_param_that"
        setpoint_name = "multi_setpoint_param_this_setpoint_set"
        loc_provider = FormatLocation(fmt=loc_fmt, record=rcd)
        c = Measure(self.p1).run(location=loc_provider)
        self.assertEqual(c.metadata['arrays'][param_name_1]['unit'], 'this unit')
        self.assertEqual(c.metadata['arrays'][param_name_1]['name'], param_name_1)
        self.assertEqual(c.metadata['arrays'][param_name_1]['label'], 'this label')
        self.assertEqual(c.metadata['arrays'][param_name_1]['is_setpoint'], False)
        self.assertEqual(c.metadata['arrays'][param_name_1]['shape'], (5,))
        assert_array_equal(getattr(c, param_name_1).ndarray, np.zeros(5))

        self.assertEqual(c.metadata['arrays'][param_name_2]['unit'], 'that unit')
        self.assertEqual(c.metadata['arrays'][param_name_2]['name'], param_name_2)
        self.assertEqual(c.metadata['arrays'][param_name_2]['label'], 'that label')
        self.assertEqual(c.metadata['arrays'][param_name_2]['is_setpoint'], False)
        self.assertEqual(c.metadata['arrays'][param_name_2]['shape'], (5,))
        assert_array_equal(getattr(c, param_name_2).ndarray, np.ones(5))

        self.assertEqual(c.metadata['arrays'][setpoint_name]['unit'],
                         'this setpointunit')
        self.assertEqual(c.metadata['arrays'][setpoint_name]['name'],
                         "multi_setpoint_param_this_setpoint")
        self.assertEqual(c.metadata['arrays'][setpoint_name]['label'],
                         'this setpoint')
        self.assertEqual(c.metadata['arrays'][setpoint_name]
                         ['is_setpoint'], True)
        self.assertEqual(c.metadata['arrays'][setpoint_name]['shape'],
                         (5,))
        assert_array_equal(getattr(c, setpoint_name).ndarray, np.linspace(5, 9, 5))



import datetime as dt
import io
import logging
import struct
import time
import xml.etree.ElementTree as ET
import zipfile as zf
from functools import partial
from typing import Any, Dict, List, Optional, Sequence, Union

import numpy as np
from broadbean.sequence import InvalidForgedSequenceError, fs_schema

from qcodes import Instrument, VisaInstrument
from qcodes import validators as vals
from qcodes.instrument.channel import ChannelList, InstrumentChannel
from qcodes.utils.validators import Validator

log = logging.getLogger(__name__)

##################################################
#
# SMALL HELPER FUNCTIONS
#


def _parse_string_response(input_str: str) -> str:
    """
    Remove quotation marks from string and return 'N/A'
    if the input is empty
    """
    output = input_str.replace('"', '')
    output = output if output else 'N/A'

    return output


##################################################
#
# MODEL DEPENDENT SETTINGS
#
# TODO: it seems that a lot of settings differ between models
# perhaps these dicts should be merged to one

_fg_path_val_map = {'5208': {'DC High BW': "DCHB",
                             'DC High Voltage': "DCHV",
                             'AC Direct': "ACD"},
                    '70001A': {'direct': 'DIR',
                               'DCamplified': 'DCAM',
                               'AC': 'AC'},
                    '70002A': {'direct': 'DIR',
                               'DCamplified': 'DCAM',
                               'AC': 'AC'},
                    '70001B': {'direct': 'DIR',
                               'DCamplified': 'DCAM',
                               'AC': 'AC'},
                    '70002B': {'direct': 'DIR',
                               'DCamplified': 'DCAM',
                               'AC': 'AC'}}

# number of markers per channel
_num_of_markers_map = {'5208': 4,
                       '70001A': 2,
                       '70002A': 2,
                       '70001B': 2,
                       '70002B': 2}

# channel resolution
_chan_resolutions = {'5208': [12, 13, 14, 15, 16],
                     '70001A': [8, 9, 10],
                     '70002A': [8, 9, 10],
                     '70001B': [8, 9, 10],
                     '70002B': [8, 9, 10]}

# channel resolution docstrings
_chan_resolution_docstrings = {'5208': "12 bit resolution allows for four "
                                       "markers, 13 bit resolution "
                                       "allows for three, etc. with 16 bit "
                                       "allowing for ZERO markers",
                               '70001A': "8 bit resolution allows for two "
                                         "markers, 9 bit resolution "
                                         "allows for one, and 10 bit "
                                         "does NOT allow for markers ",
                               '70002A': "8 bit resolution allows for two "
                                         "markers, 9 bit resolution "
                                         "allows for one, and 10 bit "
                                         "does NOT allow for markers ",
                               '70001B': "8 bit resolution allows for two "
                                         "markers, 9 bit resolution "
                                         "allows for one, and 10 bit "
                                         "does NOT allow for markers ",
                               '70002B': "8 bit resolution allows for two "
                                         "markers, 9 bit resolution "
                                         "allows for one, and 10 bit "
                                         "does NOT allow for markers "}

# channel amplitudes
_chan_amps = {'70001A': 0.5,
              '70002A': 0.5,
              '70001B': 0.5,
              '70002B': 0.5,
              '5208': 1.5}

# marker ranges
_marker_high = {'70001A': (-1.4, 1.4),
                '70002A': (-1.4, 1.4),
                '70001B': (-1.4, 1.4),
                '70002B': (-1.4, 1.4),
                '5208': (-0.5, 1.75)}
_marker_low = {'70001A': (-1.4, 1.4),
               '70002A': (-1.4, 1.4),
               '70001B': (-1.4, 1.4),
               '70002B': (-1.4, 1.4),
               '5208': (-0.3, 1.55)}


class SRValidator(Validator[float]):
    """
    Validator to validate the AWG clock sample rate
    """

    def __init__(self, awg: 'AWG70000A') -> None:
        """
        Args:
            awg: The parent instrument instance. We need this since sample
                rate validation depends on many clock settings
        """
        self.awg = awg
        if self.awg.model in ['70001A', '70001B']:
            self._internal_validator = vals.Numbers(1.49e3, 50e9)
            self._freq_multiplier = 4
        elif self.awg.model in ['70002A', '70002B']:
            self._internal_validator = vals.Numbers(1.49e3, 25e9)
            self._freq_multiplier = 2
        elif self.awg.model == '5208':
            self._internal_validator = vals.Numbers(1.49e3, 2.5e9)
        # no other models are possible, since the __init__ of
        # the AWG70000A raises an error if anything else is given

    def validate(self, value: float, context: str='') -> None:
        if 'Internal' in self.awg.clock_source():
            self._internal_validator.validate(value)
        else:
            ext_freq = self.awg.clock_external_frequency()
            # TODO: I'm not sure what the minimal allowed sample rate is
            # in this case
            validator = vals.Numbers(1.49e3, self._freq_multiplier*ext_freq)
            validator.validate(value)


class AWGChannel(InstrumentChannel):
    """
    Class to hold a channel of the AWG.
    """

    def __init__(self,  parent: Instrument, name: str, channel: int) -> None:
        """
        Args:
            parent: The Instrument instance to which the channel is
                to be attached.
            name: The name used in the DataSet
            channel: The channel number, either 1 or 2.
        """

        super().__init__(parent, name)

        self.channel = channel

        num_channels = self.root_instrument.num_channels
        self.model = self.root_instrument.model

        fg = 'function generator'

        if channel not in list(range(1, num_channels+1)):
            raise ValueError('Illegal channel value.')

        self.add_parameter('state',
                           label=f'Channel {channel} state',
                           get_cmd=f'OUTPut{channel}:STATe?',
                           set_cmd=f'OUTPut{channel}:STATe {{}}',
                           vals=vals.Ints(0, 1),
                           get_parser=int)

        ##################################################
        # FGEN PARAMETERS

        # TODO: Setting high and low will change this parameter's value
        self.add_parameter('fgen_amplitude',
                           label=f'Channel {channel} {fg} amplitude',
                           get_cmd=f'FGEN:CHANnel{channel}:AMPLitude?',
                           set_cmd=f'FGEN:CHANnel{channel}:AMPLitude {{}}',
                           unit='V',
                           vals=vals.Numbers(0, _chan_amps[self.model]),
                           get_parser=float)

        self.add_parameter('fgen_offset',
                           label=f'Channel {channel} {fg} offset',
                           get_cmd=f'FGEN:CHANnel{channel}:OFFSet?',
                           set_cmd=f'FGEN:CHANnel{channel}:OFFSet {{}}',
                           unit='V',
                           vals=vals.Numbers(0, 0.250),  # depends on ampl.
                           get_parser=float)

        self.add_parameter('fgen_frequency',
                           label=f'Channel {channel} {fg} frequency',
                           get_cmd=f'FGEN:CHANnel{channel}:FREQuency?',
                           set_cmd=partial(self._set_fgfreq, channel),
                           unit='Hz',
                           get_parser=float)

        self.add_parameter('fgen_dclevel',
                           label=f'Channel {channel} {fg} DC level',
                           get_cmd=f'FGEN:CHANnel{channel}:DCLevel?',
                           set_cmd=f'FGEN:CHANnel{channel}:DCLevel {{}}',
                           unit='V',
                           vals=vals.Numbers(-0.25, 0.25),
                           get_parser=float)

        self.add_parameter('fgen_signalpath',
                           label=f'Channel {channel} {fg} signal path',
                           set_cmd=f'FGEN:CHANnel{channel}:PATH {{}}',
                           get_cmd=f'FGEN:CHANnel{channel}:PATH?',
                           val_mapping=_fg_path_val_map[self.root_instrument.model])

        self.add_parameter('fgen_period',
                           label=f'Channel {channel} {fg} period',
                           get_cmd=f'FGEN:CHANnel{channel}:PERiod?',
                           unit='s',
                           get_parser=float)

        self.add_parameter('fgen_phase',
                           label=f'Channel {channel} {fg} phase',
                           get_cmd=f'FGEN:CHANnel{channel}:PHASe?',
                           set_cmd=f'FGEN:CHANnel{channel}:PHASe {{}}',
                           unit='degrees',
                           vals=vals.Numbers(-180, 180),
                           get_parser=float)

        self.add_parameter('fgen_symmetry',
                           label=f'Channel {channel} {fg} symmetry',
                           set_cmd=f'FGEN:CHANnel{channel}:SYMMetry {{}}',
                           get_cmd=f'FGEN:CHANnel{channel}:SYMMetry?',
                           unit='%',
                           vals=vals.Numbers(0, 100),
                           get_parser=float)

        self.add_parameter('fgen_type',
                           label=f'Channel {channel} {fg} type',
                           set_cmd=f'FGEN:CHANnel{channel}:TYPE {{}}',
                           get_cmd=f'FGEN:CHANnel{channel}:TYPE?',
                           val_mapping={'SINE': 'SINE',
                                        'SQUARE': 'SQU',
                                        'TRIANGLE': 'TRI',
                                        'NOISE': 'NOIS',
                                        'DC': 'DC',
                                        'GAUSSIAN': 'GAUSS',
                                        'EXPONENTIALRISE': 'EXPR',
                                        'EXPONENTIALDECAY': 'EXPD',
                                        'NONE': 'NONE'})

        ##################################################
        # AWG PARAMETERS

        # this command internally uses power in dBm
        # the manual claims that this command only works in AC mode
        # (OUTPut[n]:PATH is AC), but I've tested that it does what
        # one would expect in DIR mode.
        self.add_parameter(
            'awg_amplitude',
            label=f'Channel {channel} AWG peak-to-peak amplitude',
            set_cmd=f'SOURCe{channel}:VOLTage {{}}',
            get_cmd=f'SOURce{channel}:VOLTage?',
            unit='V',
            get_parser=float,
            vals=vals.Numbers(0.250, _chan_amps[self.model]))

        self.add_parameter('assigned_asset',
                           label=('Waveform/sequence assigned to '
                                  f' channel {self.channel}'),
                           get_cmd=f"SOURCE{self.channel}:CASSet?",
                           get_parser=_parse_string_response)

        # markers
        for mrk in range(1, _num_of_markers_map[self.model]+1):

            self.add_parameter(
                f'marker{mrk}_high',
                label=f'Channel {channel} marker {mrk} high level',
                set_cmd=partial(self._set_marker, channel, mrk, True),
                get_cmd=f'SOURce{channel}:MARKer{mrk}:VOLTage:HIGH?',
                unit='V',
                vals=vals.Numbers(*_marker_high[self.model]),
                get_parser=float)

            self.add_parameter(
                f'marker{mrk}_low',
                label=f'Channel {channel} marker {mrk} low level',
                set_cmd=partial(self._set_marker, channel, mrk, False),
                get_cmd=f'SOURce{channel}:MARKer{mrk}:VOLTage:LOW?',
                unit='V',
                vals=vals.Numbers(*_marker_low[self.model]),
                get_parser=float)

            self.add_parameter(
                f'marker{mrk}_waitvalue',
                label=f'Channel {channel} marker {mrk} wait state',
                set_cmd=f'OUTPut{channel}:WVALue:MARKer{mrk} {{}}',
                get_cmd=f'OUTPut{channel}:WVALue:MARKer{mrk}?',
                vals=vals.Enum('FIRST', 'LOW', 'HIGH'))

            self.add_parameter(
                name=f'marker{mrk}_stoppedvalue',
                label=f'Channel {channel} marker {mrk} stopped value',
                set_cmd=f'OUTPut{channel}:SVALue:MARKer{mrk} {{}}',
                get_cmd=f'OUTPut{channel}:SVALue:MARKer{mrk}?',
                vals=vals.Enum('OFF', 'LOW'))

        ##################################################
        # MISC.

        self.add_parameter('resolution',
                           label=f'Channel {channel} bit resolution',
                           get_cmd=f'SOURce{channel}:DAC:RESolution?',
                           set_cmd=f'SOURce{channel}:DAC:RESolution {{}}',
                           vals=vals.Enum(*_chan_resolutions[self.model]),
                           get_parser=int,
                           docstring=_chan_resolution_docstrings[self.model])

    def _set_marker(self, channel: int, marker: int,
                    high: bool, voltage: float) -> None:
        """
        Set the marker high/low value and update the low/high value
        """
        if high:
            this = 'HIGH'
            other = 'low'
        else:
            this = 'LOW'
            other = 'high'

        self.write(f'SOURce{channel}:MARKer{marker}:VOLTage:{this} {voltage}')
        self.parameters[f'marker{marker}_{other}'].get()

    def _set_fgfreq(self, channel: int, frequency: float) -> None:
        """
        Set the function generator frequency
        """
        functype = self.fgen_type.get()
        if functype in ['SINE', 'SQUARE']:
            max_freq = 12.5e9
        else:
            max_freq = 6.25e9

        # validate
        if frequency < 1 or frequency > max_freq:
            raise ValueError('Can not set channel {} frequency to {} Hz.'
                             ' Maximum frequency for function type {} is {} '
                             'Hz, minimum is 1 Hz'.format(channel, frequency,
                                                          functype, max_freq))
        else:
            self.root_instrument.write(f'FGEN:CHANnel{channel}:'
                                       f'FREQuency {frequency}')

    def setWaveform(self, name: str) -> None:
        """
        Select a waveform from the waveform list to output on this channel

        Args:
            name: The name of the waveform
        """
        if name not in self.root_instrument.waveformList:
            raise ValueError('No such waveform in the waveform list')

        self.root_instrument.write(f'SOURce{self.channel}:CASSet:WAVeform "{name}"')

    def setSequenceTrack(self, seqname: str, tracknr: int) -> None:
        """
        Assign a track from a sequence to this channel.

        Args:
            seqname: Name of the sequence in the sequence list
            tracknr: Which track to use (1 or 2)
        """

        self.root_instrument.write(f'SOURCE{self.channel}:'
                                   f'CASSet:SEQuence "{seqname}"'
                                   f', {tracknr}')


class AWG70000A(VisaInstrument):
    """
    The QCoDeS driver for Tektronix AWG70000A series AWG's.

    The drivers for AWG70001A/AWG70001B and AWG70002A/AWG70002B should be
    subclasses of this general class.
    """

    def __init__(self, name: str, address: str, num_channels: int,
                 timeout: float=10, **kwargs: Any) -> None:
        """
        Args:
            name: The name used internally by QCoDeS in the DataSet
            address: The VISA resource name of the instrument
            timeout: The VISA timeout time (in seconds)
            num_channels: Number of channels on the AWG
        """

        self.num_channels = num_channels

        super().__init__(name, address, timeout=timeout, terminator='\n',
                         **kwargs)

        # The 'model' value begins with 'AWG'
        self.model = self.IDN()['model'][3:]

        if self.model not in ['70001A', '70002A', '70001B', '70002B', '5208']:
            raise ValueError('Unknown model type: {}. Are you using '
                             'the right driver for your instrument?'
                             ''.format(self.model))

        self.add_parameter('current_directory',
                           label='Current file system directory',
                           set_cmd='MMEMory:CDIRectory "{}"',
                           get_cmd='MMEMory:CDIRectory?',
                           vals=vals.Strings())

        self.add_parameter('mode',
                           label='Instrument operation mode',
                           set_cmd='INSTrument:MODE {}',
                           get_cmd='INSTrument:MODE?',
                           vals=vals.Enum('AWG', 'FGEN'))

        ##################################################
        # Clock parameters

        self.add_parameter('sample_rate',
                           label='Clock sample rate',
                           set_cmd='CLOCk:SRATe {}',
                           get_cmd='CLOCk:SRATe?',
                           unit='Sa/s',
                           get_parser=float,
                           vals=SRValidator(self))

        self.add_parameter('clock_source',
                           label='Clock source',
                           set_cmd='CLOCk:SOURce {}',
                           get_cmd='CLOCk:SOURce?',
                           val_mapping={'Internal': 'INT',
                                        'Internal, 10 MHZ ref.': 'EFIX',
                                        'Internal, variable ref.': 'EVAR',
                                        'External': 'EXT'})

        self.add_parameter('clock_external_frequency',
                           label='External clock frequency',
                           set_cmd='CLOCk:ECLock:FREQuency {}',
                           get_cmd='CLOCk:ECLock:FREQuency?',
                           get_parser=float,
                           unit='Hz',
                           vals=vals.Numbers(6.25e9, 12.5e9))

        self.add_parameter('run_state',
                           label='Run state',
                           get_cmd='AWGControl:RSTATe?',
                           val_mapping={'Stopped': '0',
                                        'Waiting for trigger': '1',
                                        'Running': '2'})

        # We deem 2 channels too few for a channel list
        if self.num_channels > 2:
            chanlist = ChannelList(self, 'Channels', AWGChannel,
                                   snapshotable=False)

        for ch_num in range(1, num_channels+1):
            ch_name = f'ch{ch_num}'
            channel = AWGChannel(self, ch_name, ch_num)
            self.add_submodule(ch_name, channel)
            if self.num_channels > 2:
                chanlist.append(channel)

        if self.num_channels > 2:
            chanlist.lock()
            self.add_submodule('channels', chanlist)

        # Folder on the AWG where to files are uplaoded by default
        self.wfmxFileFolder = "\\Users\\OEM\\Documents"
        self.seqxFileFolder = "\\Users\\OEM\\Documents"

        self.current_directory(self.wfmxFileFolder)

        self.connect_message()

    def force_triggerA(self) -> None:
        """
        Force a trigger A event
        """
        self.write('TRIGger:IMMediate ATRigger')

    def force_triggerB(self) -> None:
        """
        Force a trigger B event
        """
        self.write('TRIGger:IMMediate BTRigger')

    def wait_for_operation_to_complete(self) -> None:
        """
        Waits for the latest issued overlapping command to finish
        """
        self.ask('*OPC?')

    def play(self, wait_for_running: bool = True,
             timeout: float = 10) -> None:
        """
        Run the AWG/Func. Gen. This command is equivalent to pressing the
        play button on the front panel.

        Args:
            wait_for_running: If True, this command is blocking while the
                instrument is getting ready to play
            timeout: The maximal time to wait for the instrument to play.
                Raises an exception is this time is reached.
        """
        self.write('AWGControl:RUN')
        if wait_for_running:
            start_time = time.perf_counter()
            running = False
            while not running:
                time.sleep(0.1)
                running = self.run_state() in ('Running',
                                               'Waiting for trigger')
                waited_for = start_time - time.perf_counter()
                if waited_for > timeout:
                    raise RuntimeError(f'Reached timeout ({timeout} s) '
                                       'while waiting for instrument to play.'
                                       ' Perhaps some waveform or sequence is'
                                       ' corrupt?')

    def stop(self) -> None:
        """
        Stop the output of the instrument. This command is equivalent to
        pressing the stop button on the front panel.
        """
        self.write('AWGControl:STOP')

    @property
    def sequenceList(self) -> List[str]:
        """
        Return the sequence list as a list of strings
        """
        # There is no SLISt:LIST command, so we do it slightly differently
        N = int(self.ask("SLISt:SIZE?"))
        slist = []
        for n in range(1, N+1):
            resp = self.ask(f"SLISt:NAME? {n}")
            resp = resp.strip()
            resp = resp.replace('"', '')
            slist.append(resp)

        return slist

    @property
    def waveformList(self) -> List[str]:
        """
        Return the waveform list as a list of strings
        """
        respstr = self.ask("WLISt:LIST?")
        respstr = respstr.strip()
        respstr = respstr.replace('"', '')
        resp = respstr.split(',')

        return resp

    def delete_sequence_from_list(self, seqname: str) -> None:
        """
        Delete the specified sequence from the sequence list

        Args:
            seqname: The name of the sequence (as it appears in the sequence
                list, not the file name) to delete
        """
        self.write(f'SLISt:SEQuence:DELete "{seqname}"')

    def clearSequenceList(self) -> None:
        """
        Clear the sequence list
        """
        self.write('SLISt:SEQuence:DELete ALL')

    def clearWaveformList(self) -> None:
        """
        Clear the waveform list
        """
        self.write('WLISt:WAVeform:DELete ALL')

    @staticmethod
    def makeWFMXFile(data: np.ndarray, amplitude: float) -> bytes:
        """
        Compose a WFMX file

        Args:
            data: A numpy array holding the data. Markers can be included.
            amplitude: The peak-to-peak amplitude (V) assumed to be set on the
                channel that will play this waveform. This information is
                needed as the waveform must be rescaled to (-1, 1) where
                -1 will correspond to the channel's min. voltage and 1 to the
                channel's max. voltage.

        Returns:
            The binary .wfmx file, ready to be sent to the instrument.
        """

        shape = np.shape(data)
        if len(shape) == 1:
            N = shape[0]
            markers_included = False
        elif len(shape) in [2, 3, 4]:
            N = shape[1]
            markers_included = True
        else:
            raise ValueError('Input data has too many dimensions!')

        wfmx_hdr_str = AWG70000A._makeWFMXFileHeader(num_samples=N,
                                                     markers_included=markers_included)
        wfmx_hdr = bytes(wfmx_hdr_str, 'ascii')
        wfmx_data = AWG70000A._makeWFMXFileBinaryData(data, amplitude)

        wfmx = wfmx_hdr

        wfmx += wfmx_data

        return wfmx

    def sendSEQXFile(self, seqx: bytes, filename: str,
                     path: Optional[str] = None) -> None:
        """
        Send a binary seqx file to the AWG's memory

        Args:
            seqx: The binary seqx file, preferably the output of
                makeSEQXFile.
            filename: The name of the file on the AWG disk, including the
                extension.
            path: The path to the directory where the file should be saved. If
                omitted, seqxFileFolder will be used.
        """
        if not path:
            path = self.seqxFileFolder

        self._sendBinaryFile(seqx, filename, path)

    def sendWFMXFile(self, wfmx: bytes, filename: str,
                     path: Optional[str] = None) -> None:
        """
        Send a binary wfmx file to the AWG's memory

        Args:
            wfmx: The binary wfmx file, preferably the output of
                makeWFMXFile.
            filename: The name of the file on the AWG disk, including the
                extension.
            path: The path to the directory where the file should be saved. If
                omitted, seqxFileFolder will be used.
        """
        if not path:
            path = self.wfmxFileFolder

        self._sendBinaryFile(wfmx, filename, path)

    def _sendBinaryFile(self, binfile: bytes, filename: str,
                        path: str, overwrite: bool = True) -> None:
        """
        Send a binary file to the AWG's mass memory (disk).

        Args:
            binfile: The binary file to send.
            filename: The name of the file on the AWG disk, including the
                extension.
            path: The path to the directory where the file should be saved.
            overwrite: If true, the file on disk gets overwritten
        """

        name_str = f'MMEMory:DATA "{filename}"'.encode('ascii')
        len_file = len(binfile)
        len_str = len(str(len_file))  # No. of digits needed to write length
        size_str = (f',#{len_str}{len_file}').encode('ascii')

        msg = name_str + size_str + binfile

        # IEEE 488.2 limit on a single write is 999,999,999 bytes
        # TODO: If this happens, we should split the file
        if len(msg) > 1e9-1:
            raise ValueError('File too large to transfer')

        self.current_directory(path)

        if overwrite:
            self.log.debug(f'Pre-deleting file {filename} at {path}')
            self.visa_handle.write(f'MMEMory:DELete "{filename}"')
            # if the file does not exist,
            # an error code -256 is put in the error queue
            resp = self.visa_handle.query(f'SYSTem:ERRor:CODE?')
            self.log.debug(f'Pre-deletion finished with return code {resp}')

        self.visa_handle.write_raw(msg)

    def loadWFMXFile(self, filename: str, path: Optional[str] = None) -> None:
        """
        Loads a wfmx from memory into the waveform list
        Only loading from the C: drive is supported

        Args:
            filename: Name of the file (with extension)
            path: Path to load from. If omitted, the default path
                (self.wfmxFileFolder) is used.
        """

        if not path:
            path = self.wfmxFileFolder

        pathstr = 'C:' + path + '\\' + filename

        self.write(f'MMEMory:OPEN "{pathstr}"')
        # the above command is overlapping, but we want a blocking command
        self.ask("*OPC?")

    def loadSEQXFile(self, filename: str, path: Optional[str] = None) -> None:
        """
        Load a seqx file from instrument disk memory. All sequences in the file
        are loaded into the sequence list.

        Args:
            filename: The name of the sequence file INCLUDING the extension
            path: Path to load from. If omitted, the default path
                (self.seqxFileFolder) is used.
        """
        if not path:
            path = self.seqxFileFolder

        pathstr = f'C:{path}\\{filename}'

        self.write(f'MMEMory:OPEN:SASSet:SEQuence "{pathstr}"')
        # the above command is overlapping, but we want a blocking command
        self.ask('*OPC?')

    @staticmethod
    def _makeWFMXFileHeader(num_samples: int,
                            markers_included: bool) -> str:
        """
        Compiles a valid XML header for a .wfmx file
        There might be behaviour we can't capture

        We always use 9 digits for the number of header character
        """
        offsetdigits = 9

        if not isinstance(num_samples, int):
            raise ValueError('num_samples must be of type int.')

        if num_samples < 2400:
            raise ValueError('num_samples must be at least 2400.')

        # form the timestamp string
        timezone = time.timezone
        tz_m, _ = divmod(timezone, 60)  # returns (minutes, seconds)
        tz_h, tz_m = divmod(tz_m, 60)
        if np.sign(tz_h) == -1:
            signstr = '-'
            tz_h *= -1
        else:
            signstr = '+'
        timestr = dt.datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3]
        timestr += signstr
        timestr += f'{tz_h:02.0f}:{tz_m:02.0f}'

        hdr = ET.Element('DataFile', attrib={'offset': '0'*offsetdigits,
                                             'version': '0.1'})
        dsc = ET.SubElement(hdr, 'DataSetsCollection')
        dsc.set("xmlns", "http://www.tektronix.com")
        dsc.set("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
        dsc.set("xsi:schemaLocation", (r"http://www.tektronix.com file:///" +
                                       r"C:\Program%20Files\Tektronix\AWG70000" +
                                       r"\AWG\Schemas\awgDataSets.xsd"))
        datasets = ET.SubElement(dsc, 'DataSets')
        datasets.set('version', '1')
        datasets.set("xmlns", "http://www.tektronix.com")

        # Description of the data
        datadesc = ET.SubElement(datasets, 'DataDescription')
        temp_elem = ET.SubElement(datadesc, 'NumberSamples')
        temp_elem.text = f'{num_samples:d}'
        temp_elem = ET.SubElement(datadesc, 'SamplesType')
        temp_elem.text = 'AWGWaveformSample'
        temp_elem = ET.SubElement(datadesc, 'MarkersIncluded')
        temp_elem.text = (f'{markers_included}').lower()
        temp_elem = ET.SubElement(datadesc, 'NumberFormat')
        temp_elem.text = 'Single'
        temp_elem = ET.SubElement(datadesc, 'Endian')
        temp_elem.text = 'Little'
        temp_elem = ET.SubElement(datadesc, 'Timestamp')
        temp_elem.text = timestr

        # Product specific information
        prodspec = ET.SubElement(datasets, 'ProductSpecific')
        prodspec.set('name', '')
        temp_elem = ET.SubElement(prodspec, 'ReccSamplingRate')
        temp_elem.set('units', 'Hz')
        temp_elem.text = 'NaN'
        temp_elem = ET.SubElement(prodspec, 'ReccAmplitude')
        temp_elem.set('units', 'Volts')
        temp_elem.text = 'NaN'
        temp_elem = ET.SubElement(prodspec, 'ReccOffset')
        temp_elem.set('units', 'Volts')
        temp_elem.text = 'NaN'
        temp_elem = ET.SubElement(prodspec, 'SerialNumber')
        temp_elem = ET.SubElement(prodspec, 'SoftwareVersion')
        temp_elem.text = '1.0.0917'
        temp_elem = ET.SubElement(prodspec, 'UserNotes')
        temp_elem = ET.SubElement(prodspec, 'OriginalBitDepth')
        temp_elem.text = 'Floating'
        temp_elem = ET.SubElement(prodspec, 'Thumbnail')
        temp_elem = ET.SubElement(prodspec, 'CreatorProperties',
                          attrib={'name': ''})
        temp_elem = ET.SubElement(hdr, 'Setup')

        xmlstr = ET.tostring(hdr, encoding='unicode')
        xmlstr = xmlstr.replace('><', '>\r\n<')

        # As the final step, count the length of the header and write this
        # in the DataFile tag attribute 'offset'

        xmlstr = xmlstr.replace('0'*offsetdigits,
                                '{num:0{pad}d}'.format(num=len(xmlstr),
                                                       pad=offsetdigits))

        return xmlstr

    @staticmethod
    def _makeWFMXFileBinaryData(data: np.ndarray, amplitude: float) -> bytes:
        """
        For the binary part.

        Note that currently only zero markers or two markers are supported;
        one-marker data will break.

        Args:
            data: Either a shape (N,) array with only a waveform or
                a shape (M, N) array with waveform, marker1, marker2, marker3, i.e.
                data = np.array([wfm, m1, ...]). The waveform data is assumed
                to be in V.
            amplitude: The peak-to-peak amplitude (V) assumed to be set on the
                channel that will play this waveform. This information is
                needed as the waveform must be rescaled to (-1, 1) where
                -1 will correspond to the channel's min. voltage and 1 to the
                channel's max. voltage.
        """

        channel_max = amplitude/2
        channel_min = -amplitude/2

        shape = np.shape(data)

        if len(shape) == 1:
            N = shape[0]
            binary_marker = b''
            wfm = data
        else:
            N = shape[1]
            M = shape[0]
            wfm = data[0, :]
            markers = data[1, :]
            for i in range(1, M-1):
                markers += data[i+1, :] * (2**i)
            markers = markers.astype(int)
            fmt = N*'B'  # endian-ness doesn't matter for one byte
            binary_marker = struct.pack(fmt, *markers)

        if wfm.max() > channel_max or wfm.min() < channel_min:
            log.warning('Waveform exceeds specified channel range.'
                        ' The resulting waveform will be clipped. '
                        'Waveform min.: {} (V), waveform max.: {} (V),'
                        'Channel min.: {} (V), channel max.: {} (V)'
                        ''.format(wfm.min(), wfm.max(), channel_min,
                                  channel_max))

        # the data must be such that channel_max becomes 1 and
        # channel_min becomes -1
        scale = 2/amplitude
        wfm = wfm*scale

        # TODO: Is this a fast method?
        fmt = '<' + N*'f'
        binary_wfm = struct.pack(fmt, *wfm)
        binary_out = binary_wfm + binary_marker

        return binary_out

    @staticmethod
    def make_SEQX_from_forged_sequence(
            seq: Dict[int, Dict[Any, Any]],
            amplitudes: List[float],
            seqname: str,
            channel_mapping: Optional[Dict[Union[str, int], int]] = None
    ) -> bytes:
        """
        Make a .seqx from a forged broadbean sequence.
        Supports subsequences.

        Args:
            seq: The output of broadbean's Sequence.forge()
            amplitudes: A list of the AWG channels' voltage amplitudes.
                The first entry is ch1 etc.
            channel_mapping: A mapping from what the channel is called
                in the broadbean sequence to the integer describing the
                physical channel it should be assigned to.
            seqname: The name that the sequence will have in the AWG's
                sequence list. Used for loading the sequence.

        Returns:
            The binary .seqx file contents. Can be sent directly to the
                instrument or saved on disk.
        """

        try:
            fs_schema.validate(seq)
        except Exception as e:
            raise InvalidForgedSequenceError(e)

        chan_list: List[Union[str, int]] = []
        for pos1 in seq.keys():
            for pos2 in seq[pos1]['content'].keys():
                for ch in seq[pos1]['content'][pos2]['data'].keys():
                    if ch not in chan_list:
                        chan_list.append(ch)

        if channel_mapping is None:
            channel_mapping = {ch: ch_ind+1
                               for ch_ind, ch in enumerate(chan_list)}

        if len(set(chan_list)) != len(amplitudes):
            raise ValueError('Incorrect number of amplitudes provided.')

        if set(chan_list) != set(channel_mapping.keys()):
            raise ValueError(f'Invalid channel_mapping. The sequence has '
                             f'channels {set(chan_list)}, but the '
                             'channel_mapping maps from the channels '
                             f'{set(channel_mapping.keys())}')

        if set(channel_mapping.values()) != set(range(1, 1+len(chan_list))):
            raise ValueError('Invalid channel_mapping. Must map onto '
                             f'{list(range(1, 1+len(chan_list)))}')

        ##########
        # STEP 1:
        # Make all .wfmx files

        wfmx_files: List[bytes] = []
        wfmx_filenames: List[str] = []

        for pos1 in seq.keys():
            for pos2 in seq[pos1]['content'].keys():
                for ch, data in seq[pos1]['content'][pos2]['data'].items():
                    wfm = data['wfm']

                    markerdata = []
                    for mkey in ['m1', 'm2', 'm3', 'm4']:
                        if mkey in data.keys():
                            markerdata.append(data.get(mkey))
                    wfm_data = np.stack((wfm, *markerdata))

                    awgchan = channel_mapping[ch]
                    wfmx = AWG70000A.makeWFMXFile(wfm_data,
                                                  amplitudes[awgchan-1])
                    wfmx_files.append(wfmx)
                    wfmx_filenames.append(f'wfm_{pos1}_{pos2}_{awgchan}')

        ##########
        # STEP 2:
        # Make all subsequence .sml files

        log.debug(f'Waveforms done: {wfmx_filenames}')

        subseqsml_files: List[str] = []
        subseqsml_filenames: List[str] = []

        for pos1 in seq.keys():
            if seq[pos1]['type'] == 'subsequence':

                ss_wfm_names: List[List[str]] = []

                # we need to "flatten" all the individual dicts of element
                # sequence options into one dict of lists of sequencing options
                # and we must also provide default values if nothing
                # is specified
                seqings: List[Dict[str, int]] = []
                for pos2 in (seq[pos1]['content'].keys()):
                    pos_seqs = seq[pos1]['content'][pos2]['sequencing']
                    pos_seqs['twait'] = pos_seqs.get('twait', 0)
                    pos_seqs['nrep'] = pos_seqs.get('nrep', 1)
                    pos_seqs['jump_input'] = pos_seqs.get('jump_input', 0)
                    pos_seqs['jump_target'] = pos_seqs.get('jump_target', 0)
                    pos_seqs['goto'] = pos_seqs.get('goto', 0)
                    seqings.append(pos_seqs)

                    ss_wfm_names.append([n for n in wfmx_filenames
                                         if f'wfm_{pos1}_{pos2}' in n])

                seqing = {k: [d[k] for d in seqings]
                          for k in seqings[0].keys()}

                subseqname = f'subsequence_{pos1}'

                log.debug(f'Subsequence waveform names: {ss_wfm_names}')

                subseqsml = AWG70000A._makeSMLFile(trig_waits=seqing['twait'],
                                                   nreps=seqing['nrep'],
                                                   event_jumps=seqing['jump_input'],
                                                   event_jump_to=seqing['jump_target'],
                                                   go_to=seqing['goto'],
                                                   elem_names=ss_wfm_names,
                                                   seqname=subseqname,
                                                   chans=len(channel_mapping))

                subseqsml_files.append(subseqsml)
                subseqsml_filenames.append(f'{subseqname}')

        ##########
        # STEP 3:
        # Make the main .sml file

        asset_names: List[List[str]] = []
        seqings = []
        subseq_positions: List[int] = []
        for pos1 in seq.keys():
            pos_seqs = seq[pos1]['sequencing']

            pos_seqs['twait'] = pos_seqs.get('twait', 0)
            pos_seqs['nrep'] = pos_seqs.get('nrep', 1)
            pos_seqs['jump_input'] = pos_seqs.get('jump_input', 0)
            pos_seqs['jump_target'] = pos_seqs.get('jump_target', 0)
            pos_seqs['goto'] = pos_seqs.get('goto', 0)
            seqings.append(pos_seqs)
            if seq[pos1]['type'] == 'subsequence':
                subseq_positions.append(pos1)
                asset_names.append([sn for sn in subseqsml_filenames
                                    if f'_{pos1}' in sn])
            else:
                asset_names.append([wn for wn in wfmx_filenames
                                    if f'wfm_{pos1}' in wn])
        seqing = {k: [d[k] for d in seqings] for k in seqings[0].keys()}

        log.debug(f'Assets for SML file: {asset_names}')

        mainseqname = seqname
        mainseqsml = AWG70000A._makeSMLFile(trig_waits=seqing['twait'],
                                            nreps=seqing['nrep'],
                                            event_jumps=seqing['jump_input'],
                                            event_jump_to=seqing['jump_target'],
                                            go_to=seqing['goto'],
                                            elem_names=asset_names,
                                            seqname=mainseqname,
                                            chans=len(channel_mapping),
                                            subseq_positions=subseq_positions)

        ##########
        # STEP 4:
        # Build the .seqx file

        user_file = b''
        setup_file = AWG70000A._makeSetupFile(mainseqname)

        buffer = io.BytesIO()

        zipfile = zf.ZipFile(buffer, mode='a')
        for ssn, ssf in zip(subseqsml_filenames, subseqsml_files):
            zipfile.writestr(f'Sequences/{ssn}.sml', ssf)
        zipfile.writestr(f'Sequences/{mainseqname}.sml', mainseqsml)

        for (name, wfile) in zip(wfmx_filenames, wfmx_files):
            zipfile.writestr(f'Waveforms/{name}.wfmx', wfile)

        zipfile.writestr('setup.xml', setup_file)
        zipfile.writestr('userNotes.txt', user_file)
        zipfile.close()

        buffer.seek(0)
        seqx = buffer.getvalue()
        buffer.close()

        return seqx

    @staticmethod
    def makeSEQXFile(trig_waits: Sequence[int],
                     nreps: Sequence[int],
                     event_jumps: Sequence[int],
                     event_jump_to: Sequence[int],
                     go_to: Sequence[int],
                     wfms: Sequence[Sequence[np.ndarray]],
                     amplitudes: Sequence[float],
                     seqname: str,
                     flags: Optional[Sequence[Sequence[Sequence[int]]]] = None
                     ) -> bytes:
        """
        Make a full .seqx file (bundle)
        A .seqx file can presumably hold several sequences, but for now
        we support only packing a single sequence

        For a single sequence, a .seqx file is a bundle of two files and
        two folders:

        /Sequences
            sequence.sml

        /Waveforms
            wfm1.wfmx
            wfm2.wfmx
            ...

        setup.xml
        userNotes.txt

        Args:
            trig_waits: Wait for a trigger? If yes, you must specify the
                trigger input. 0 for off, 1 for 'TrigA', 2 for 'TrigB',
                3 for 'Internal'.
            nreps: No. of repetitions. 0 corresponds to infinite.
            event_jumps: Jump when event triggered? If yes, you must specify
                the trigger input. 0 for off, 1 for 'TrigA', 2 for 'TrigB',
                3 for 'Internal'.
            event_jump_to: Jump target in case of event. 1-indexed,
                0 means next. Must be specified for all elements.
            go_to: Which element to play next. 1-indexed, 0 means next.
            wfms: numpy arrays describing each waveform plus two markers,
                packed like np.array([wfm, m1, m2]). These numpy arrays
                are then again packed in lists according to:
                [[wfmch1pos1, wfmch1pos2, ...], [wfmch2pos1, ...], ...]
            amplitudes: The peak-to-peak amplitude in V of the channels, i.e.
                a list [ch1_amp, ch2_amp].
            seqname: The name of the sequence. This name will appear in the
                sequence list. Note that all spaces are converted to '_'
            flags: Flags for the auxiliary outputs. 0 for 'No change', 1 for
                'High', 2 for 'Low', 3 for 'Toggle', or 4 for 'Pulse'. 4 flags
                [A, B, C, D] for every channel in every element, packed like:
                [[ch1pos1, ch1pos2, ...], [ch2pos1, ...], ...]
                If omitted, no flags will be set.

        Returns:
            The binary .seqx file, ready to be sent to the instrument.
        """

        # input sanitising to avoid spaces in filenames
        seqname = seqname.replace(' ', '_')

        (chans, elms) = (len(wfms), len(wfms[0]))
        wfm_names = [[f'wfmch{ch}pos{el}' for ch in range(1, chans+1)]
                     for el in range(1, elms+1)]

        # generate wfmx files for the waveforms
        flat_wfmxs = [] # type: List[bytes]
        for amplitude, wfm_lst in zip(amplitudes, wfms):
            flat_wfmxs += [AWG70000A.makeWFMXFile(wfm, amplitude)
                           for wfm in wfm_lst]

        # This unfortunately assumes no subsequences
        flat_wfm_names = list(np.reshape(np.array(wfm_names).transpose(),
                                         (chans*elms,)))

        sml_file = AWG70000A._makeSMLFile(trig_waits, nreps,
                                          event_jumps, event_jump_to,
                                          go_to, wfm_names,
                                          seqname,
                                          chans, flags=flags)

        user_file = b''
        setup_file = AWG70000A._makeSetupFile(seqname)

        buffer = io.BytesIO()

        zipfile = zf.ZipFile(buffer, mode='a')
        zipfile.writestr(f'Sequences/{seqname}.sml', sml_file)

        for (name, wfile) in zip(flat_wfm_names, flat_wfmxs):
            zipfile.writestr(f'Waveforms/{name}.wfmx', wfile)

        zipfile.writestr('setup.xml', setup_file)
        zipfile.writestr('userNotes.txt', user_file)
        zipfile.close()

        buffer.seek(0)
        seqx = buffer.getvalue()
        buffer.close()

        return seqx

    @staticmethod
    def _makeSetupFile(sequence: str) -> str:
        """
        Make a setup.xml file.

        Args:
            sequence: The name of the main sequence

        Returns:
            The setup file as a string
        """
        head = ET.Element('RSAPersist')
        head.set('version', '0.1')
        temp_elem = ET.SubElement(head, 'Application')
        temp_elem.text = 'Pascal'
        temp_elem = ET.SubElement(head, 'MainSequence')
        temp_elem.text = sequence
        prodspec = ET.SubElement(head, 'ProductSpecific')
        prodspec.set('name', 'AWG70002A')
        temp_elem = ET.SubElement(prodspec, 'SerialNumber')
        temp_elem.text = 'B020397'
        temp_elem = ET.SubElement(prodspec, 'SoftwareVersion')
        temp_elem.text = '5.3.0128.0'
        temp_elem = ET.SubElement(prodspec, 'CreatorProperties')
        temp_elem.set('name', '')

        xmlstr = ET.tostring(head, encoding='unicode')
        xmlstr = xmlstr.replace('><', '>\r\n<')

        return xmlstr

    @staticmethod
    def _makeSMLFile(trig_waits: Sequence[int],
                     nreps: Sequence[int],
                     event_jumps: Sequence[int],
                     event_jump_to: Sequence[int],
                     go_to: Sequence[int],
                     elem_names: Sequence[Sequence[str]],
                     seqname: str,
                     chans: int,
                     subseq_positions: Sequence[int] = (),
                     flags: Optional[Sequence[Sequence[Sequence[int]]]] = None
                     ) -> str:
        """
        Make an xml file describing a sequence.

        Args:
            trig_waits: Wait for a trigger? If yes, you must specify the
                trigger input. 0 for off, 1 for 'TrigA', 2 for 'TrigB',
                3 for 'Internal'.
            nreps: No. of repetitions. 0 corresponds to infinite.
            event_jumps: Jump when event triggered? If yes, you must specify
                the trigger input. 0 for off, 1 for 'TrigA', 2 for 'TrigB',
                3 for 'Internal'.
            event_jump_to: Jump target in case of event. 1-indexed,
                0 means next. Must be specified for all elements.
            go_to: Which element to play next. 1-indexed, 0 means next.
            elem_names: The waveforms/subsequences to use. Should be packed
                like:
                [[wfmpos1ch1, wfmpos1ch2, ...],
                 [subseqpos2],
                 [wfmpos3ch1, wfmpos3ch2, ...], ...]
            seqname: The name of the sequence. This name will appear in
                the sequence list of the instrument.
            chans: The number of channels. Can not be inferred in the case
                of a sequence containing only subsequences, so must be provided
                up front.
            subseq_positions: The positions (step numbers) occupied by
                subsequences
            flags: Flags for the auxiliary outputs. 0 for 'No change', 1 for
                'High', 2 for 'Low', 3 for 'Toggle', or 4 for 'Pulse'. 4 flags
                [A, B, C, D] for every channel in every element, packed like:
                [[ch1pos1, ch1pos2, ...], [ch2pos1, ...], ...]
                If omitted, no flags will be set.

        Returns:
            A str containing the file contents, to be saved as an .sml file
        """

        offsetdigits = 9

        waitinputs = {0: 'None', 1: 'TrigA', 2: 'TrigB', 3: 'Internal'}
        eventinputs = {0: 'None', 1: 'TrigA', 2: 'TrigB', 3: 'Internal'}
        flaginputs = {0:'NoChange', 1:'High', 2:'Low', 3:'Toggle', 4:'Pulse'}

        inputlsts = [trig_waits, nreps, event_jump_to, go_to]
        lstlens = [len(lst) for lst in inputlsts]
        if lstlens.count(lstlens[0]) != len(lstlens):
            raise ValueError('All input lists must have the same length!')

        if lstlens[0] == 0:
            raise ValueError('Received empty sequence option lengths!')

        if lstlens[0] != len(elem_names):
            raise ValueError('Mismatch between number of waveforms and'
                             ' number of sequencing steps.')

        N = lstlens[0]

        # form the timestamp string
        timezone = time.timezone
        tz_m, _ = divmod(timezone, 60)
        tz_h, tz_m = divmod(tz_m, 60)
        if np.sign(tz_h) == -1:
            signstr = '-'
            tz_h *= -1
        else:
            signstr = '+'
        timestr = dt.datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3]
        timestr += signstr
        timestr += f'{tz_h:02.0f}:{tz_m:02.0f}'

        datafile = ET.Element('DataFile', attrib={'offset': '0'*offsetdigits,
                                                  'version': '0.1'})
        dsc = ET.SubElement(datafile, 'DataSetsCollection')
        dsc.set("xmlns", "http://www.tektronix.com")
        dsc.set("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
        dsc.set("xsi:schemaLocation", (r"http://www.tektronix.com file:///" +
                                       r"C:\Program%20Files\Tektronix\AWG70000" +
                                       r"\AWG\Schemas\awgSeqDataSets.xsd"))
        datasets = ET.SubElement(dsc, 'DataSets')
        datasets.set('version', '1')
        datasets.set("xmlns", "http://www.tektronix.com")

        # Description of the data
        datadesc = ET.SubElement(datasets, 'DataDescription')
        temp_elem = ET.SubElement(datadesc, 'SequenceName')
        temp_elem.text = seqname
        temp_elem = ET.SubElement(datadesc, 'Timestamp')
        temp_elem.text = timestr
        temp_elem = ET.SubElement(datadesc, 'JumpTiming')
        temp_elem.text = 'JumpImmed'  # TODO: What does this control?
        temp_elem = ET.SubElement(datadesc, 'RecSampleRate')
        temp_elem.text = 'NaN'
        temp_elem = ET.SubElement(datadesc, 'RepeatFlag')
        temp_elem.text = 'false'
        temp_elem = ET.SubElement(datadesc, 'PatternJumpTable')
        temp_elem.set('Enabled', 'false')
        temp_elem.set('Count', '65536')
        steps = ET.SubElement(datadesc, 'Steps')
        steps.set('StepCount', f'{N:d}')
        steps.set('TrackCount', f'{chans:d}')

        for n in range(1, N+1):
            step = ET.SubElement(steps, 'Step')
            temp_elem = ET.SubElement(step, 'StepNumber')
            temp_elem.text = f'{n:d}'
            # repetitions
            rep = ET.SubElement(step, 'Repeat')
            repcount = ET.SubElement(step, 'RepeatCount')
            if nreps[n-1] == 0:
                rep.text = 'Infinite'
                repcount.text = '1'
            elif nreps[n-1] == 1:
                rep.text = 'Once'
                repcount.text = '1'
            else:
                rep.text = "RepeatCount"
                repcount.text = f"{nreps[n-1]:d}"
            # trigger wait
            temp_elem = ET.SubElement(step, 'WaitInput')
            temp_elem.text = waitinputs[trig_waits[n-1]]
            # event jump
            temp_elem = ET.SubElement(step, 'EventJumpInput')
            temp_elem.text = eventinputs[event_jumps[n-1]]
            jumpto = ET.SubElement(step, 'EventJumpTo')
            jumpstep = ET.SubElement(step, 'EventJumpToStep')
            if event_jump_to[n-1] == 0:
                jumpto.text = 'Next'
                jumpstep.text = '1'
            else:
                jumpto.text = "StepIndex"
                jumpstep.text = f"{event_jump_to[n-1]:d}"
            # Go to
            goto = ET.SubElement(step, 'GoTo')
            gotostep = ET.SubElement(step, 'GoToStep')
            if go_to[n-1] == 0:
                goto.text = 'Next'
                gotostep.text = '1'
            else:
                goto.text = "StepIndex"
                gotostep.text = f"{go_to[n-1]:d}"

            assets = ET.SubElement(step, 'Assets')
            for assetname in elem_names[n-1]:
                asset = ET.SubElement(assets, 'Asset')
                temp_elem = ET.SubElement(asset, 'AssetName')
                temp_elem.text = assetname
                temp_elem = ET.SubElement(asset, 'AssetType')
                if n in subseq_positions:
                    temp_elem.text = 'Sequence'
                else:
                    temp_elem.text = 'Waveform'

            # convert flag settings to strings
            flags_list = ET.SubElement(step, 'Flags')
            for chan in range(chans):
                flagset = ET.SubElement(flags_list, 'FlagSet')
                for flgind, flg in enumerate(['A', 'B', 'C', 'D']):
                    temp_elem = ET.SubElement(flagset, 'Flag')
                    temp_elem.set('name', flg)
                    if flags is None:
                        # no flags were passed to the function
                        temp_elem.text = 'NoChange'
                    else:
                        temp_elem.text = flaginputs[flags[chan][n-1][flgind]]

        temp_elem = ET.SubElement(datasets, 'ProductSpecific')
        temp_elem.set('name', '')
        temp_elem = ET.SubElement(datafile, 'Setup')

        # the tostring() call takes roughly 75% of the total
        # time spent in this function. Can we speed up things?
        # perhaps we should use lxml?
        xmlstr = ET.tostring(datafile, encoding='unicode')
        xmlstr = xmlstr.replace('><', '>\r\n<')

        # As the final step, count the length of the header and write this
        # in the DataFile tag attribute 'offset'

        xmlstr = xmlstr.replace('0'*offsetdigits,
                                '{num:0{pad}d}'.format(num=len(xmlstr),
                                                       pad=offsetdigits))

        return xmlstr



import json
import sys

from tqdm import tqdm

from qcodes.dataset.descriptions.versioning.v0 import InterDependencies
from qcodes.dataset.sqlite.connection import ConnectionPlus, atomic, \
    atomic_transaction
from qcodes.dataset.sqlite.queries import get_run_description, \
    update_run_description
from qcodes.dataset.sqlite.query_helpers import one


def upgrade_5_to_6(conn: ConnectionPlus) -> None:
    """
    Perform the upgrade from version 5 to version 6.

    The upgrade ensures that the runs_description has a top-level entry
    called 'version'. Note that version changes of the runs_description will
    not be tracked as schema upgrades.
    """
    no_of_runs_query = "SELECT max(run_id) FROM runs"
    no_of_runs = one(atomic_transaction(conn, no_of_runs_query), 'max(run_id)')
    no_of_runs = no_of_runs or 0

    # If one run fails, we want the whole upgrade to roll back, hence the
    # entire upgrade is one atomic transaction

    with atomic(conn) as conn:
        pbar = tqdm(range(1, no_of_runs+1), file=sys.stdout)
        pbar.set_description("Upgrading database; v5 -> v6")

        empty_idps_ser = InterDependencies()._to_dict()

        for run_id in pbar:
            json_str = get_run_description(conn, run_id)
            if json_str is None:
                new_json = json.dumps({'version': 0,
                                       'interdependencies': empty_idps_ser})
            else:
                ser = json.loads(json_str)
                new_ser = {'version': 0}  # let 'version' be the first entry
                new_ser['interdependencies'] = ser['interdependencies']
                new_json = json.dumps(new_ser)
            update_run_description(conn, run_id, new_json)
